/* $Header: CM_funcs.c,v 6.4 93/10/13 16:47:16 rkm Locked $
 *------------------------------------------------------------*
 * DESCRIPTION
 *	CMPSL functions to make error handling a little easier.
 */


#include <stdio.h>
#include <sys/file.h>
#include <sys/types.h>
#include <general.h>

#define MAX_FILES	32
#define FILE_LEN	256

static char file_names[MAX_FILES][FILE_LEN];

char *
_CM_file_name(fs)
/*------------------------------------------------------------*
 * RETURN
 *	the name of the file opened on fs with CM_fopen();
 */
FILE *fs;
{
    return file_names[fileno(fs) % MAX_FILES];
}


FILE *
_CM_fopen (file, mode, srcfile, srcline)
char *file;		/* file name */
char *mode;		/* mode */
char *srcfile;		/* name of calling source file */
int32 srcline;		/* line no. in calling source file */
{
    FILE *fs = fopen (file, mode);

    if (fs == NULL) {
	fprintf (stdout, "%s(%d): fopen(%s,%s) failed\n",
		 srcfile, srcline, file, mode);
	perror ("fopen");
	exit (-1);
    }

    strncpy (file_names[fileno(fs)%MAX_FILES], file, FILE_LEN);
    return (fs);
}

FILE *
_CM_fopenp (dirl, file, mode, srcfile, srcline)
char *dirl;		/* dir list */
char *file;		/* file name */
char *mode;		/* mode */
char *srcfile;		/* name of source file */
int32   srcline;		/* line no. in source file */
{
    char buffer[2048];
    FILE *fs;

    sprintf (buffer, "%s/%s", dirl, file);
    fs = (FILE *) fopen (buffer, mode);

    if (fs == NULL) {
	fprintf (stdout, "%s(%d): fopen(%s,%s) failed\n", srcfile, srcline,
		 buffer, mode);
	perror ("fopen");
	exit (-1);
    }

    strncpy (file_names[fileno(fs)%MAX_FILES], file, FILE_LEN);
    return (fs);
}

char *_CM_calloc (cnt, size, file, line)
int32 cnt;		/* Number of items */
int32 size;		/* Size of items */
char *file;		/* name of calling file */
int32 line;		/* line no. file */
{
    char *ret;

    if (cnt == 0)
	return 0;

    ret = (char *) calloc ((size_t)cnt, (size_t)size);
    if (ret == 0) {
	fprintf (stdout, "%s(%d): calloc(%d,%d) failed\n", file, line, cnt, size);
	exit (-1);
    }
    return (ret);
}


char *_CM_2dcalloc (rcnt, ccnt, size, srcfile, srcline)
/*------------------------------------------------------------*
 * DESCRIPTION - allocate row pointers and data in one chunk
 */
int32 rcnt;		/* Number of rows */
int32 ccnt;		/* Number of cols */
int32 size;		/* Size of items */
char *srcfile;		/* source file name */
int32   srcline;		/* source file line no. */
{
    char 	*ret;
    caddr_t 	*rowPtr;
    int32	r;

    if ((rcnt == 0) || (ccnt == 0))
	return 0;

    ret = (char *) calloc ((size_t)(rcnt * ccnt * size) +
			   rcnt * sizeof(caddr_t), 1);
    rowPtr = (caddr_t *) ret;

    if (ret == 0) {
	fprintf (stdout, "%s(%d): CM_2dcalloc(%d,%d,%d) failed\n", srcfile, srcline,
	  	 rcnt, ccnt, size);
	exit (-1);
    }

    for (r = 0; r < rcnt; r++)
	rowPtr[r] = ret + (rcnt * sizeof(caddr_t)) + (r * ccnt * size);

    return (ret);
}

char *_CM_3dcalloc (lcnt, rcnt, ccnt, size, srcfile, srcline)
/*------------------------------------------------------------*
 * DESCRIPTION - allocate row pointers and data in one chunk
 */
int32 lcnt;		/* Number of levels */
int32 rcnt;		/* Number of rows */
int32 ccnt;		/* Number of cols */
int32 size;		/* Size of items */
char *srcfile;		/* source file name */
int32   srcline;		/* source file line no. */
{
    char 	*ret;
    caddr_t 	*rowPtr;
    caddr_t 	*lvlPtr;
    int32	r, l;

    ret = (char *) calloc ((size_t)(lcnt * rcnt * ccnt * size) +
			   (size_t)(lcnt * rcnt * sizeof(caddr_t)) + 
			   (size_t)(lcnt * sizeof(caddr_t)), 1);
    rowPtr = (caddr_t *) ret;
    lvlPtr = (caddr_t *) ret;

    if (ret == 0) {
	fprintf (stdout, "%s(%d): CM_3dcalloc(%d,%d,%d) failed\n",
		 srcfile, srcline, rcnt, ccnt, size);
	exit (-1);
    }

    for (l = 0; l < lcnt; l++) {
	lvlPtr[l] = ret + (lcnt * sizeof(caddr_t)) + (rcnt * sizeof(caddr_t) * l);
	rowPtr = (caddr_t *) lvlPtr[l];
	for (r = 0; r < rcnt; r++) {
	    rowPtr[r] = ret + (lcnt * sizeof(caddr_t)) +
		        (lcnt * rcnt * sizeof(caddr_t)) +
			(l * rcnt * ccnt * size) +
			(r * ccnt * size);
	}
    }

    return (ret);
}


char *_CM_recalloc (ptr, cnt, size, srcfile, srcline)
char *ptr;		/* original pointer */
int32 cnt;		/* Number of items */
int32 size;		/* Size of items */
char *srcfile;		/* source file name */
int32   srcline;		/* source file line no. */
{
    char *ret;

    if (ptr == 0)
	ret = (char *) calloc ((size_t)cnt, (size_t)size);
    else
	ret = (char *) realloc (ptr, (size_t)size * (size_t)cnt);

    if (ret == 0) {
	fprintf (stdout, "%s(%d): recalloc(0x%X,%d,%d) failed\n", srcfile, srcline, 
		 ptr, cnt, size);
	exit (-1);
    }
    return (ret);
}
