/* era.c - for ERA (Error Region Analysis)
 *
 *-----------------------------------------------------------------------*
 * HISTORY
 *
 * 20-Jul-95  L L Chase (chase@cs.cmu.edu) at LM95/Johns Hopkins
 *      Extracted error analysis component from CMU lattice analysis toolkit.
 *
 *   era.c compares two lattice paths looking for error regions
 *   and writes a description of what it finds to a human-readable
 *   file. 
 */

/* Notes: This file is currently being modified A LOT.
 * Changes left to make:
 *       test
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <strings.h>
#include <math.h>
#include "lat.h"
/*#include <log.h>*/
#include "comp.h"
#include "error.h"
#include "general.h"
#include <tcl.h>
#include <tk.h>
#include <CM_macros.h>

extern int sent_err;

int frame_tol;

int prune_regs_ac,no_prune_regs_ac;
int prune_regs_lm,no_prune_regs_lm;
int prune_regs_both,no_prune_regs_both;
float prune_both_ratio_lmtoa,no_prune_both_ratio_lmtoa;

int ref_tot_bigger;
int hyp_tot_bigger;
int num_error_regions;   /* total number of error regions in utt set */
int num_unk_error_regions = 0;
int num_ref_phones = 0;
int num_hyp_phones = 0;
int num_ref_errreg_phones = 0;
int num_hyp_errreg_phones = 0;
int num_hyp_incorrect_phones = 0;

force_out_t hyp;                                 /* decoder output contents */
force_out_t ref;                                 /* transcript contents */

int32 ref_err_st_seg,ref_err_en_seg; /* for pairing code */
int32 hyp_err_st_seg,hyp_err_en_seg; /* for pairing code */
int32 ref_ctl_seg;                     /* for pairing code */
int32 hyp_ctl_seg;                     /* for pairing code */
int error_mode,error_found;
int32 last_hyp_rejoin_seg;
/* int num_utt_err_regions;          num error regions in utterance */
int num_correct_phones = 0;


int ldbug = 0;
int NumCommonWords = 0;

/* Temp small change */

/*extern dictT *WordDict;*/  /* from CMU version */

void make_region_data(force_out_t *refptr, force_out_t *hypptr, int32 ress, int32 rees, int32 hess, int32 hees, err_region_t *err_region, int subregion, int tol);

void print0_err_region(force_out_t *refptr, force_out_t *hypptr, FILE *outfp, int32 ress,int32 rees,int32 hess,int32 hees, err_region_t *err_region, int print_iter_num);

void print1_err_region(force_out_t *refptr, force_out_t *hypptr, FILE *outfp, int32 ress,int32 rees,int32 hess,int32 hees, err_region_t *err_region, int print_iter_num);

void print2_err_region(force_out_t *refptr, force_out_t *hypptr, FILE *outfp, int32 ress,int32 rees,int32 hess,int32 hees, err_region_t *err_region, int print_iter_num);

void print3_err_region(force_out_t *refptr, force_out_t *hypptr, FILE *outfp, int32 ress,int32 rees,int32 hess,int32 hees, err_region_t *err_region, int print_iter_num);

void print_visual_err_region(force_out_t *refptr, force_out_t *hypptr, char *uttid, char *tkdir, int32 ress,int32 rees,int32 hess,int32 hees, err_region_t *err_region, int print_iter_num,int reg_no,int norm, int sc_ratio);

err_region_t *make_new_err_region(void);

void make_sub_region(force_out_t *refptr, force_out_t *hypptr, int32 rss, int32 res,int32 hss,int32 hes,err_region_t *sub_err_reg,err_region_t *err_region);

void stripname(phname,basename)
char *phname,*basename;
{
  int i;
  int phlen;
  int blen = 0;

  phlen = strlen(phname);
  for (i=0;i<phlen;i++)  {
    if (phname[i] != '(')  {
      basename[i] = phname[i];
      blen++;
    }
    else  {
      break;
    }
  }
  basename[blen] = '\0';
}

void do_error_reg_analysis(ref_latpath, hyp_latpath, uttid, tkdir, window, tol, norm, sc_ratio)
lat_path_t *ref_latpath,*hyp_latpath;
char *uttid,*tkdir;
int window,tol,norm;
int sc_ratio;
{
  int i,j,k;
  int count;
  int num_ref_segs,num_hyp_segs;
  char lm_str[1024];

  int err_regions_left = MAX_REGS;
  int realloc_num = 2;
  err_region_t **err_region;
  int print_iter_num;
  int subregion;

  int num_phones = 0;

  int num_utt_err_regions = 0;
  int num_printed_err_regions = 0;

  /*FILE *datafp;*/
  FILE *outfp;

  /*datafp = CM_fopen (datafile, "a");*/
#if 0
  outfp = fopen (outfile, "a");
  if (outfp == NULL)  {
    printf("ERROR:  Unable to open output file %s.\n",outfile);
    exit(-1);
  }
#endif

#if 0
  printf("Initializing Visual ERA (visera).\n");
  fflush(stdout);
  init_ui();
  printf("Testing window creation.\n");
  fflush(stdout);
  ui_errwin_create(2500,100,2500);
#endif

#if 0
  fprintf(outfp,"UTTERANCE: %s\n",uttid);
  fflush(outfp);
#endif

  align_ref_and_hyp_latpaths(ref_latpath,hyp_latpath,uttid,tkdir);

  ref.num_segs = ref_latpath->num_segs;
  num_ref_segs = ref_latpath->num_segs;
  /*fprintf(outfp,"REF:");*/
  for (i=0;i<num_ref_segs;i++)  {
    lm_str[0] = '\0';
    strcpy(ref.segs[i].word,ref_latpath->seg_ptrs[i]->word);
    ref.segs[i].start_f = ref_latpath->seg_ptrs[i]->sf;
    ref.segs[i].end_f = ref_latpath->seg_ptrs[i]->ef;
    ref.segs[i].acoust_s = ref_latpath->seg_ptrs[i]->ac_score;
    ref.segs[i].lm_s = ref_latpath->LM_ptrs[i]->lm_score;
    strcpy(ref.segs[i].lm_s_src,ref_latpath->LM_ptrs[i]->lm_src);
    /*fprintf(outfp," %s",ref.segs[i].word);*/
    num_phones = ref_latpath->seg_ptrs[i]->phone_seq.num_phones;
    ref.segs[i].phone_seq.num_phones = num_phones;
    for (k=0;k<num_phones;k++)  {
      strcpy(ref.segs[i].phone_seq.phones[k].name,
	     ref_latpath->seg_ptrs[i]->phone_seq.phones[k].name);
      stripname(ref.segs[i].phone_seq.phones[k].name,
		ref.segs[i].phone_seq.phones[k].basename);
      ref.segs[i].phone_seq.phones[k].sf =
	ref_latpath->seg_ptrs[i]->phone_seq.phones[k].sf;
      ref.segs[i].phone_seq.phones[k].ef =
	ref_latpath->seg_ptrs[i]->phone_seq.phones[k].ef;
      ref.segs[i].phone_seq.phones[k].ac_score =
	ref_latpath->seg_ptrs[i]->phone_seq.phones[k].ac_score;
    }
  }
  /*fprintf(outfp,"\n");*/

  /*printf("Reference segments in error analysis:\n");*/
  /*for (i=0;i<num_ref_segs;i++)  {
    printf("[%d]%s <%d,%d> {ac:%li,lm:%li(%s)}\n",
	   i,
	   ref.segs[i].word,
	   ref.segs[i].start_f,
	   ref.segs[i].end_f,
	   ref.segs[i].acoust_s,
	   ref.segs[i].lm_s,
	   ref.segs[i].lm_s_src);
  }
  */

  hyp.num_segs = hyp_latpath->num_segs;
  num_hyp_segs = hyp_latpath->num_segs;

  /*fprintf(outfp,"HYP:");*/

  for (i=0;i<num_hyp_segs;i++)  {
    lm_str[0] = '\0';
    strcpy(hyp.segs[i].word,hyp_latpath->seg_ptrs[i]->word);
    hyp.segs[i].start_f = hyp_latpath->seg_ptrs[i]->sf;
    hyp.segs[i].end_f = hyp_latpath->seg_ptrs[i]->ef;
    hyp.segs[i].acoust_s = hyp_latpath->seg_ptrs[i]->ac_score;
    hyp.segs[i].lm_s = hyp_latpath->LM_ptrs[i]->lm_score;
    strcpy(hyp.segs[i].lm_s_src,hyp_latpath->LM_ptrs[i]->lm_src);
    /*fprintf(outfp," %s",hyp.segs[i].word);*/
    num_phones = hyp_latpath->seg_ptrs[i]->phone_seq.num_phones;
    hyp.segs[i].phone_seq.num_phones = num_phones;
    for (k=0;k<num_phones;k++)  {
      strcpy(hyp.segs[i].phone_seq.phones[k].name,
	     hyp_latpath->seg_ptrs[i]->phone_seq.phones[k].name);
      stripname(hyp.segs[i].phone_seq.phones[k].name,
		hyp.segs[i].phone_seq.phones[k].basename);
      hyp.segs[i].phone_seq.phones[k].sf =
	hyp_latpath->seg_ptrs[i]->phone_seq.phones[k].sf;
      hyp.segs[i].phone_seq.phones[k].ef =
	hyp_latpath->seg_ptrs[i]->phone_seq.phones[k].ef;
      hyp.segs[i].phone_seq.phones[k].ac_score =
	hyp_latpath->seg_ptrs[i]->phone_seq.phones[k].ac_score;
    }
  }
  /*fprintf(outfp,"\n");*/

  /*printf("Hypothesis segments in error analysis:\n");*/
  /*for (i=0;i<num_hyp_segs;i++)  {
    printf("[%d]%s <%d,%d> {ac:%li,lm:%li(%s)}\n",
	   i,
	   hyp.segs[i].word,
	   hyp.segs[i].start_f,
	   hyp.segs[i].end_f,
	   hyp.segs[i].acoust_s,
	   hyp.segs[i].lm_s,
	   hyp.segs[i].lm_s_src);
  }
  */

  /* Tag all phones as OK initially -- will be updated later */
  /* First, ref phones */

  for (i=0;i<num_ref_segs;i++)  {
    num_phones = ref.segs[i].phone_seq.num_phones;
    for (k=0;k<num_phones;k++)  {
      ref.segs[i].phone_seq.phones[k].err_reg = 0;
      ref.segs[i].phone_seq.phones[k].correct = 1;
    }
  }
  /*  Now hyp phones */
  
  for (i=0;i<num_hyp_segs;i++)  {
    num_phones = hyp.segs[i].phone_seq.num_phones;
    for (k=0;k<num_phones;k++)  {
      hyp.segs[i].phone_seq.phones[k].err_reg = 0;
      hyp.segs[i].phone_seq.phones[k].correct = 1;
    }
  }

  /* Begin looking for and processing error regions */

  ref_err_st_seg = 0;
  ref_err_en_seg = 0;
  hyp_err_st_seg = 0;
  hyp_err_en_seg = 0;

  ref_ctl_seg = 0;
  hyp_ctl_seg = 0;
  
  error_mode = 0;
  error_found = 0;
  last_hyp_rejoin_seg = -1;

  frame_tol = tol;

  /*printf("num_ref_segs is %d\n",num_ref_segs);*/
  /*printf("num_hyp_segs is %d\n",num_hyp_segs);*/
  while ((ref_ctl_seg < num_ref_segs) && (hyp_ctl_seg < num_hyp_segs))  {
    /* skip all segments that are the same in ref and hyp */
    /*printf("ref_ctl_seg is %d\n",ref_ctl_seg);
    printf("hyp_ctl_seg is %d\n",hyp_ctl_seg);
    printf("num_ref_segs is %d\n",num_ref_segs);
    printf("num_hyp_segs is %d\n",num_hyp_segs);
    printf("x");
    fflush(stdout);*/
    /*fflush(datafp);*/
    for (j=ref_ctl_seg; j<num_ref_segs; j++)  {
      /*
      printf("j,starting at ref_ctl_seg up to num_ref_segs, is %d\n",j);
      printf("ref: %s (%d,%d)\n",ref.segs[j].word,ref.segs[j].start_f,ref.segs[j].end_f);
      printf("hyp: %s (%d,%d)\n",hyp.segs[hyp_ctl_seg].word,hyp.segs[hyp_ctl_seg].start_f,hyp.segs[hyp_ctl_seg].end_f);
      fflush(stdout);      
      */
      if (!match_force_out_seg(&ref.segs[j],&hyp.segs[hyp_ctl_seg]))  {
	/*printf("no match at ref seg %d, starting error region\n",j);*/
	/*printf("Reference segments in error analysis:\n");*/
	/*fflush(stdout);*/
	/*for (i=0;i<num_ref_segs;i++)  {
	  printf("[%d]%s(%d)<%d,%d> {ac:%d,lm:%d(%s)}\n",
		 i,
		 ref.segs[i].word,
		 ref.segs[i].wid,
		 ref.segs[i].start_f,
		 ref.segs[i].end_f,
		 ref.segs[i].acoust_s,
		 ref.segs[i].lm_s,
		 ref.segs[i].lm_s_src);
	}
	*/
	fflush(stdout);
	ref_err_st_seg = j;
	hyp_err_st_seg = hyp_ctl_seg;
	error_mode = 1;
	error_found = 1;
	break;
	/* deal with an error region now */
      }
      /*printf("incrementing hyp_ctl_seg,ref_ctl_seg\n");
      fflush(stdout);*/
      hyp_ctl_seg++;
      ref_ctl_seg++;
    }
    /* find the end of the error region -- it may be the end of
       the utterance.  {ref,hyp}_err_en_seg are set to the
       final errorful segments.  the segment pair after these,
       if it exists, is guaranteed to be a good match */
    if (error_mode)  {
      /*printf("in error mode, window is %d\n",window);
      fflush(stdout);*/
      if (window == 0)  {
	find_rejoin0(&ref,&hyp,ref_err_st_seg,hyp_err_st_seg,
		     &ref_err_en_seg,&hyp_err_en_seg);
      }
      else if (window == 1)  {
	find_rejoin1(&ref,&hyp,ref_err_st_seg,hyp_err_st_seg,
		     &ref_err_en_seg,&hyp_err_en_seg);
      }
      else if (window == 2)  {
	find_rejoin2(&ref,&hyp,ref_err_st_seg,hyp_err_st_seg,
		     &ref_err_en_seg,&hyp_err_en_seg);
      }
      else  {
	printf("Error!  Window size %d not supported.  Exiting.\n",window);
	exit(-2);
      }
      /*printf("out of find_rejoin, ref_err_en_seg is %d, hyp_err_en_seg is %d\n",
	     ref_err_en_seg,hyp_err_en_seg);
      fflush(stdout);*/
      num_error_regions++;
      num_utt_err_regions++;
      /*  this function builds the region and subregions */
      /*printf("entering make_region_data\n");
      fflush(stdout);*/
      if ((hyp_err_en_seg-hyp_err_st_seg) > 
	  (ref_err_en_seg-ref_err_st_seg))  {
	/*printf("Error region %d contains %d words.\n",
	       num_error_regions,
	       (hyp_err_en_seg-hyp_err_st_seg+1));
	       */
      }
      else  {
	/*printf("Error region %d contains %d words.\n",
	       num_error_regions,
	       (ref_err_en_seg-ref_err_st_seg+1));
	       */
      }
      fflush(stdout);

      /*printf("\nError regions left = %d\n", err_regions_left);*/
      if (err_regions_left == MAX_REGS)  {
        err_region = (err_region_t **)malloc(MAX_REGS*sizeof(err_region_t *));
        if (!err_region)  {
          printf("WARNING: Malloc does not have enough space for error region array.\n");
          exit (-1);
        }
        else  {
          /*printf("Successfully allocated error region space (10 places).\n");*/
	}
      }

      if (err_regions_left == 0)  {

         /*printf("err_regions_left == 0 so I am reallocating memory.\n");*/
         err_region = (err_region_t **) realloc(err_region,
                          MAX_REGS*realloc_num*sizeof(err_region_t *));
        realloc_num++;
        /*printf("Realloc num = %d.\n", realloc_num);*/
        if (!err_region)  {
          printf("WARNING: Realloc does not have enough space for error region array.\n");
          exit(-1);
        }
        else {
          /*printf("Reallocated error region space.\n");*/
	}
      }

      /*printf("Calling make_new_err_region.\n");*/

      err_region[num_utt_err_regions - 1] = make_new_err_region(); 
      
      /*printf("Out of make_new_error_region.\n");*/

      err_regions_left  = err_regions_left - 1;

      ref.num_regs += 1;
      hyp.num_regs += 1;

/* Stuff error region with known data */

      err_region[num_utt_err_regions - 1]->sf = ref.segs[ref_err_st_seg].start_f;
      err_region[num_utt_err_regions - 1]->ef = ref.segs[ref_err_en_seg].end_f;
      err_region[num_utt_err_regions - 1]->ref = ref_latpath;
      err_region[num_utt_err_regions - 1]->hyp = hyp_latpath;

      /*printf("Calling make_region_data from main routine.\n");*/

      subregion = 0;
      make_region_data(&ref,&hyp,ref_err_st_seg,ref_err_en_seg,
		       hyp_err_st_seg,hyp_err_en_seg,
		       err_region[num_utt_err_regions - 1],subregion,
		       tol);

      /*printf("Made it out of make_region_data ok.\n");*/
      /*printf("ready to print\n");*/

      fflush(stdout); 

      /*printf("Setting print_iter_num now.\n");*/

      print_iter_num = 1;

      /*printf("Evaluating which print function to call now.\n");*/
      fflush(stdout);

#if 0
      if (verbosity == 0) {
	/*printf("Calling print0_err_region routine.\n");*/
	fflush(stdout);
	print0_err_region(&ref,&hyp,outfp, ref_err_st_seg,ref_err_en_seg,hyp_err_st_seg,hyp_err_en_seg,err_region[num_utt_err_regions - 1],print_iter_num);
      }
      else if (verbosity == 1) {
	print1_err_region(&ref,&hyp,outfp, ref_err_st_seg,ref_err_en_seg,hyp_err_st_seg,hyp_err_en_seg,err_region[num_utt_err_regions - 1], print_iter_num);
      }
      else if (verbosity == 2) {
	print2_err_region(&ref,&hyp,outfp, ref_err_st_seg,ref_err_en_seg,hyp_err_st_seg,hyp_err_en_seg,err_region[num_utt_err_regions - 1],print_iter_num);
      }
      else if (verbosity == 3) {
	print3_err_region(&ref,&hyp,outfp, ref_err_st_seg,ref_err_en_seg,hyp_err_st_seg,hyp_err_en_seg,err_region[num_utt_err_regions - 1],print_iter_num);
      }
      else  {
	printf("WARNING: Unknown verbosity level: %d.\n",verbosity);
      }
#endif
      fflush(stdout);
      if (notSILs(err_region[num_utt_err_regions - 1],
		  ref_err_st_seg,
		  ref_err_en_seg,
		  hyp_err_st_seg,
		  hyp_err_en_seg,
		  &ref,
		  &hyp))  {
	/*printf("Calling visual display.\n");*/
	num_printed_err_regions++;
	print_visual_err_region(&ref,
				&hyp,
				uttid, 
				tkdir,
				ref_err_st_seg,
				ref_err_en_seg,
				hyp_err_st_seg,
				hyp_err_en_seg,
				err_region[num_utt_err_regions - 1],
				print_iter_num,
				num_printed_err_regions,
				norm,
				sc_ratio);
      }
      /*printf("printing complete\n");*/
      fflush(stdout);
      /*fflush(datafp);*/
      last_hyp_rejoin_seg = hyp_err_en_seg;
      /*printf("last_hyp_rejoin_seg is %d\n",last_hyp_rejoin_seg);*/
      fflush(stdout);
      error_mode = 0;
      ref_ctl_seg = ref_err_en_seg + 1;
      hyp_ctl_seg = hyp_err_en_seg + 1;
      /*printf("(2)ref_ctl_seg is %d\n",ref_ctl_seg);
      printf("(2)hyp_ctl_seg is %d\n",hyp_ctl_seg);
      */
      fflush(stdout);
    }
    else  {
      /*printf("not error mode\n");*/
      /*printf("(3)ref_ctl_seg is %d\n",ref_ctl_seg);
      printf("(3)hyp_ctl_seg is %d\n",hyp_ctl_seg);
      */
      fflush(stdout);
    }
  }
  if (!error_found)  {
    /*printf("Errcmp:  No errors found.\n");*/
    num_utt_err_regions = 0;
  }
  printf("ErrorRegions %d\n", num_printed_err_regions);
  printf("UttErrors %d\n",sent_err);
  fflush(stdout);

  num_phones = 0;
  num_ref_phones = 0;
  num_ref_errreg_phones = 0;
  /* run through REF and HYP phones, count and report statistics */
  for (i=0;i<num_ref_segs;i++)  {
    num_phones = ref.segs[i].phone_seq.num_phones;
    num_ref_phones = num_ref_phones + num_phones;
    for (j=0;j<num_phones;j++)  {
      if (ref.segs[i].phone_seq.phones[j].err_reg)  {
	num_ref_errreg_phones++;
      }
    }
  }
  printf("RefPhones %d\n",num_ref_phones);
  printf("RefErrRegPhones %d\n",num_ref_errreg_phones);
  fflush(stdout);

  num_correct_phones = num_ref_phones - num_ref_errreg_phones;

  num_phones = 0;
  num_hyp_phones = 0;
  num_hyp_errreg_phones = 0;
  num_hyp_incorrect_phones = 0;

  for (i=0;i<num_hyp_segs;i++)  {
    num_phones = hyp.segs[i].phone_seq.num_phones;
    num_hyp_phones = num_hyp_phones + num_phones;
    for (j=0;j<num_phones;j++)  {
      if (hyp.segs[i].phone_seq.phones[j].err_reg)  {
	num_hyp_errreg_phones++;
      }
      if (!(hyp.segs[i].phone_seq.phones[j].correct))  {
	num_hyp_incorrect_phones++;
      }
    }
  }
  printf("HypPhones %d\n",num_hyp_phones);
  printf("HypErrRegPhones %d\n",num_hyp_errreg_phones);
  printf("HypIncorrectPhones %d\n",num_hyp_incorrect_phones);
  fflush(stdout);

#if 0
  if (num_correct_phones != (num_hyp_phones - num_hyp_errreg_phones))  {
    printf("WARNING:  %d correct phones but %d - %d = %d hyp phones!\n",
	   num_correct_phones,
	   num_hyp_phones,
	   num_hyp_errreg_phones,
	   (num_hyp_phones - num_hyp_errreg_phones));
    count = 0;
    for (i=0;i<num_ref_segs;i++)  {
      num_phones = ref.segs[i].phone_seq.num_phones;
      for (j=0;j<num_phones;j++)  {
	count++;
	printf("REF[%d] is %s, marked %d\n",
	       count,
	       ref.segs[i].phone_seq.phones[j].name,
	       ref.segs[i].phone_seq.phones[j].err_reg);
	fflush(stdout);
      }
    }
    count = 0;
    for (i=0;i<num_hyp_segs;i++)  {
      num_phones = hyp.segs[i].phone_seq.num_phones;
      for (j=0;j<num_phones;j++)  {
	count++;
	printf("HYP[%d] is %s, marked %d\n",
	       count,
	       hyp.segs[i].phone_seq.phones[j].name,
	       hyp.segs[i].phone_seq.phones[j].err_reg);
	fflush(stdout);
      }
    }
  }

#endif

  /*ref.num_regs = num_utt_err_regions;    */  
  ref.num_regs = num_printed_err_regions;   
  /*hyp.num_regs = num_utt_err_regions;      */
  hyp.num_regs = num_printed_err_regions;

  fflush(stdout);
}


/*******************************/
/*  Builds database elements for an error region, writes them to a file
    for later consumption.  (Need read routines to match, will consider
    writing binary data structures instead of ASCII later.)

    Builds a region and its subregions.  Writes them to dfp file.
*/

void make_region_data(refptr,hypptr,ress,rees,hess,hees,err_region,subregion,tol)
force_out_t *refptr,*hypptr;
int32 ress,rees,hess,hees;
err_region_t *err_region;
int subregion;
int tol;
{
  int32 i,j,ref_st_sub,ref_en_sub,hyp_st_sub,hyp_en_sub;
  int32 num_ref,num_hyp;
  int32 ref_tot_LM,hyp_tot_LM,ref_tot_AC,hyp_tot_AC;
  int32 ref_tot,hyp_tot;
  int32 ref_reg_LM,hyp_reg_LM,ref_reg_AC,hyp_reg_AC;
  int32 ref_reg_tot,hyp_reg_tot;
  int num_sub_regions = 0;
  int frame_tolerance;
  int sub_region_found = 0;
  int sub_err_regions_left = MAX_SUBS;
  int realloc_num = 2;
  int last_call = 0;

  frame_tolerance = tol;
  /*printf("Using frame tolerance of %d.\n",frame_tolerance);*/
  /*printf("Entered make_region_data routine.\n");*/

  num_ref = rees - ress + 1;
  num_hyp = hees - hess + 1;

  err_region->ref_data.st_seg = ress;
  err_region->ref_data.en_seg = rees;
  err_region->hyp_data.st_seg = hess;
  err_region->hyp_data.en_seg = hees;

  /*  calculate info for top level error region, stuff what's
      not already there */

  ref_tot_LM=0;
  ref_tot_AC=0;
  hyp_tot_LM=0;
  hyp_tot_AC=0;
  ref_tot = 0;
  hyp_tot = 0;

  /* printf("Calculating region data.\n");*/

  /* these calculations are for the entire error region */
  /* calculate lm stuff for ref segs */
  for (i=0;i<num_ref;i++)  {
    ref_tot_LM = ref_tot_LM + refptr->segs[ress+i].lm_s;
  }
  /* calculate lm stuff for hyp segs */
  for (i=0;i<num_hyp;i++)  {
    hyp_tot_LM = hyp_tot_LM + hypptr->segs[hess+i].lm_s;
  }
  /* calculate acoustic stuff for ref segs */
  for (i=0;i<num_ref;i++)  {
    ref_tot_AC = ref_tot_AC + refptr->segs[ress+i].acoust_s;
  }
  /* calculate acoustic stuff for hyp segs */
  for (i=0;i<num_hyp;i++)  {
    hyp_tot_AC = hyp_tot_AC + hypptr->segs[hess+i].acoust_s;
  }

  ref_tot = ref_tot_LM+ref_tot_AC;
  hyp_tot = hyp_tot_LM+hyp_tot_AC;

  /*stuff the values */
  err_region->ref_data.tot_s = ref_tot;
  err_region->ref_data.lm_s = ref_tot_LM;
  err_region->ref_data.acoust_s = ref_tot_AC;
  err_region->hyp_data.tot_s = hyp_tot;
  err_region->hyp_data.lm_s = hyp_tot_LM;
  err_region->hyp_data.acoust_s = hyp_tot_AC;

  /* was this region a result of a pruning error? */
  /*  (need a higher level concept -- which one to put the info in?) */
  if (ref_tot>hyp_tot)  {
    err_region->ref_data.pruning = 1;
    err_region->hyp_data.pruning = 1;
  }
  else  {
    err_region->ref_data.pruning = 0;
    err_region->hyp_data.pruning = 0;
  }


  if (!subregion) {

    /*  now identify, calculate, and stuff subregion information */
    /* traverse the reference segments, identifying how they should
       be grouped and which hyp segments belong to the groups */

    /*printf("Entering subregion detection area. \n");*/

    /*   ref_en_sub = ress-1;
         hyp_en_sub = hess-1;
         printf("ref_en_sub is %d\n",ref_en_sub);
         printf("hyp_en_sub is %d\n",hyp_en_sub);	  
    */     

    /*printf("ress is %d\n",ress);
    printf("hess is %d\n",hess);
    printf("rees is %d\n",rees);
    printf("hees is %d\n",hees);
    */
  
    ref_st_sub = ress;
    hyp_st_sub = hess;


    while ( ((ref_st_sub<rees) || (hyp_st_sub<hees)) || last_call)  {
      /*printf("z\n");*/
      fflush(stdout);
      sub_region_found = 0;
    
      for (i=ref_st_sub;i<=rees;i++) {
        if (sub_region_found)
          break;

        for (j=hyp_st_sub;j<=hees;j++)  {
	  /*printf("Calling tolerance_match i == %d, j == %d\n",i,j);*/
          fflush(stdout);
	  if (tolerance_match(refptr,hypptr,i,j,frame_tolerance))  {
	    /*printf("found (end_f) match at i == %d j == %d\n",i,j);
            printf("ref_st_sub = %d, ress = %d\n", ref_st_sub, ress);
            printf("i = %d, rees = %d\n", i, rees);
	    printf("hyp_st_sub = %d, hess = %d\n", hyp_st_sub, hess);
            printf("j = %d, hees = %d\n", j, hees);
	    fflush(stdout);
	    */
            if ((ref_st_sub == ress) && (i == rees) &&
                (hyp_st_sub == hess) && (j == hees)) {
              /* Have just found the original error region - not a sub region */
              /*printf("Found the original error region - breaking out.\n");*/
	      sub_region_found = 1;
              ref_st_sub = rees;
              hyp_st_sub = hees;
              break; 
            }

	    /*printf("Found a sub region.\n");*/
          
            sub_region_found = 1;  

            /* DO SUBREGION PROCESSING NOW */
            /* Create subregion error space if this is the first subregion
            found. If there is no more space available, reallocate some
            new space. */

            /*printf("Sub error regions left = %d \n", sub_err_regions_left);*/

            if (sub_err_regions_left == MAX_SUBS)  {
              err_region->child = (err_region_t **)malloc(MAX_SUBS*sizeof(err_region_t *));
              /*printf("Allocated sub error region space.\n");*/
              if (!err_region->child)  {
                printf("WARNING: Malloc does not have enough space for sub error region array.\n");
                exit (-1);
              }
            }  
     
            if (sub_err_regions_left == 0)  {
              err_region->child = (err_region_t **) realloc(err_region->child,
                                   MAX_SUBS*realloc_num*sizeof(err_region_t *));
              /*printf("Reallocated sub error region space.\n");*/
              realloc_num++; 
              if (!err_region->child)  {
                printf("WARNING: Realloc does not have enough space for sub error region array.\n");
                exit(-1);
              }
            }

            /*printf("Creating a new sub region.\n");*/
	    /*printf("num_sub_regions = %d\n",num_sub_regions);*/
            err_region->child[num_sub_regions] = make_new_err_region(); 

            sub_err_regions_left = sub_err_regions_left - 1;
         
            /* Stuff sub error region with known data */

            err_region->child[num_sub_regions]->parent = err_region;
            err_region->child[num_sub_regions]->sf = ref.segs[ref_st_sub].start_f;
            err_region->child[num_sub_regions]->ef = ref.segs[i].end_f;

            /* Is there something to put here?
            err_region->child[num_sub_regions]->ref = ref_latpath;
            err_region->child[num_sub_regions]->hyp = hyp_latpath;
            */

            /*printf("Calling make_region_data as a recursive sub region call.\n");*/

	    subregion = 1;
            make_region_data(refptr,hypptr,ref_st_sub,i,
	                     hyp_st_sub,j,
                             err_region->child[num_sub_regions],
                             subregion,tol);
		  	
	    /*printf("Exited make_region_data as a recursive call.\n");*/

	    ref_st_sub = i + 1;
	    hyp_st_sub = j + 1;
  
	    num_sub_regions++;

            if ((ref_st_sub == rees) && (hyp_st_sub == hees))
              last_call = 1;
	    else
	      last_call = 0;

           break;

	  }  /* if */
        }  /* for */
      }  /* for */
    }  /* while */

    err_region->num_children = num_sub_regions;
/*    printf("ref region found (%d %d)A%d LM%d P%d with %d subs\n",
	   err_region->ref_data.st_seg,
	   err_region->ref_data.en_seg,
	   err_region->ref_data.acoust_s,
	   err_region->ref_data.lm_s,
	   err_region->ref_data.pruning,
	   err_region->num_children);
    printf("hyp region found (%d %d)A%d LM%d P%d with %d subs\n",
	   err_region->hyp_data.st_seg,
	   err_region->hyp_data.en_seg,
	   err_region->hyp_data.acoust_s,
	   err_region->hyp_data.lm_s,
	   err_region->hyp_data.pruning,
	   err_region->num_children);
*/
  }  /* if */
}

/*******************************/  
/*  looks at end frames of segments to see if they are within
    <frame_tolerance> of each other
*/
int tolerance_match(refptr,hypptr,i,j,fr_tol)
force_out_t *refptr,*hypptr;
int32 i,j;
int fr_tol;
{
  /*printf("ref.segs[%d].end_f == %d, hyp.segs[%d].end_f == %d\n",
	 i,refptr->segs[i].end_f,
	 j,hypptr->segs[j].end_f);
  printf("frame tolerance is %d\n",fr_tol);
  fflush(stdout);
  */
  if (abs((int)(refptr->segs[i].end_f - hypptr->segs[j].end_f)) 
        <= fr_tol)  {
    /*printf("yes\n");
    fflush(stdout);*/
    return(1);
  }
  else  {
    /*printf("no\n");
    fflush(stdout);*/
    return(0);
  }
}
  
/*******************************/
/*  create and fill in a subregion delimited by the arguments */

/* RIGHT NOW THIS ROUTINE IS NOT CALLED AT ALL!!!! */

void make_sub_region(refptr,hypptr,rss,res,hss,hes,sub_err_reg,err_region)
force_out_t *refptr,*hypptr;
int32 rss,res,hss,hes;
err_region_t *sub_err_reg, *err_region;

{
  int32 ref_reg_LM=0;
  int32 ref_reg_AC=0;
  int32 hyp_reg_LM=0;
  int32 hyp_reg_AC=0;
  int32 ref_reg_tot = 0;
  int32 hyp_reg_tot = 0;
  int32 i;

  /* these calculations are for the subregion */  
  for (i=rss;i<=res;i++)  {
    ref_reg_AC += refptr->segs[i].acoust_s;
    ref_reg_LM += refptr->segs[i].lm_s;
    ref_reg_tot = ref_reg_tot + ref_reg_AC + ref_reg_LM;
  }
  for (i=hss;i<=hes;i++)  {
    hyp_reg_AC += hypptr->segs[i].acoust_s;
    hyp_reg_LM += hypptr->segs[i].lm_s;
    hyp_reg_tot = hyp_reg_tot + hyp_reg_AC + hyp_reg_LM;
  }

  sub_err_reg->ref_data.acoust_s = ref_reg_AC;
  sub_err_reg->ref_data.lm_s = ref_reg_LM;
  sub_err_reg->ref_data.tot_s = ref_reg_tot;
  sub_err_reg->ref_data.st_seg = rss;
  sub_err_reg->ref_data.en_seg = res;

  sub_err_reg->hyp_data.acoust_s = hyp_reg_AC;
  sub_err_reg->hyp_data.lm_s = hyp_reg_LM;
  sub_err_reg->hyp_data.tot_s = hyp_reg_tot;
  sub_err_reg->hyp_data.st_seg = hss;
  sub_err_reg->hyp_data.en_seg = hes;
  
  /* one word to one word cases*/
  if ((rss==res) && (hss==hes))  {
    /* subset with same words and segmentations */
    if (match_force_out_seg(&refptr->segs[rss],&hypptr->segs[hss]))  {
      /*  LM diffs only at end of error region?  class 3 */
      if (((err_region->ref_data.en_seg - res)<2) ||
	  ((err_region->hyp_data.en_seg - hes)<2))  {
	sub_err_reg->ref_data.align_type = 3;
	sub_err_reg->hyp_data.align_type = 3;
      }
      /*  LM diffs only but in middle of region somewhere  class 2 */
      else  {
	sub_err_reg->ref_data.align_type = 2;
	sub_err_reg->hyp_data.align_type = 2;
      }
    }
    /* subset with different segmentations or words -- class 1 */
    else  {
      sub_err_reg->ref_data.align_type = 1;
      sub_err_reg->hyp_data.align_type = 1;
    }
  }
  /* multiple words in region (ref and/or hyp) */
  else  {
    /* more segments in ref sub_region than in hyp */
    if ((sub_err_reg->ref_data.en_seg -	sub_err_reg->ref_data.st_seg) >
	(sub_err_reg->hyp_data.en_seg -	sub_err_reg->hyp_data.st_seg))  {
      sub_err_reg->ref_data.align_type = 5;
      sub_err_reg->hyp_data.align_type = 5;
    }
    /* more segments in hyp sub_region than in ref */
    else if ((sub_err_reg->ref_data.en_seg - sub_err_reg->ref_data.st_seg) <
	     (sub_err_reg->hyp_data.en_seg - sub_err_reg->hyp_data.st_seg))  {
      sub_err_reg->ref_data.align_type = 6;
      sub_err_reg->hyp_data.align_type = 6;
    }
    /* same number of segments in ref and hyp sub_regions, just > 1 */
    else if ((sub_err_reg->ref_data.en_seg - sub_err_reg->ref_data.st_seg) ==
	     (sub_err_reg->hyp_data.en_seg - sub_err_reg->hyp_data.st_seg))  {
      sub_err_reg->ref_data.align_type = 7;
      sub_err_reg->hyp_data.align_type = 7;
    }
    else  {
      printf("ERROR in assigning align_type.\n");
    }
  }

/*  printf("   ref subseg (%d %d) A%d L%d align %d\n",
	 sub_err_reg->ref_data.st_seg,
	 sub_err_reg->ref_data.en_seg,
	 sub_err_reg->ref_data.acoust_s,
	 sub_err_reg->ref_data.lm_s,
	 sub_err_reg->ref_data.align_type);
  printf("   hyp subseg (%d %d) A%d L%d align %d\n",
	 sub_err_reg->hyp_data.st_seg,
	 sub_err_reg->hyp_data.en_seg,
	 sub_err_reg->hyp_data.acoust_s,
	 sub_err_reg->hyp_data.lm_s,
	 sub_err_reg->hyp_data.align_type);
*/

/* What?
  return(sub_err_reg->ref_data.align_type);
*/

}


/*******************************/
/*  looks for two healthy segment matches, both of which have "real" LM
    components (no sil or filler words can count towards the two).
    Sets index_i to be the end segment of ref, index_j to be the end segment
    of hyp in the error region (if an end is found, otherwise the initial
    values are not changed).

    replaces:  
        if ((match_force_out_seg(&refptr->segs[i],&hypptr->segs[j])) &&
	  (match_force_out_seg(&refptr->segs[i+1],&hypptr->segs[j+1])))  {
*/

/* needs to be updated and switched on a flag */
int end_of_error_region(refptr,hypptr,index_i,index_j)
force_out_t *refptr,*hypptr;
int32 *index_i,*index_j;

{
  int32 num_real_matches_found = 0;
  int32 k,l;

  l = *index_j;
  k = *index_i;
  while ((l<(hypptr->num_segs)-1) && (k<(refptr->num_segs)-1))  {
    if (match_force_out_seg(&refptr->segs[k],&hypptr->segs[l]))  {
      if ((strcmp(refptr->segs[k].lm_s_src,"SIL")) &&
	  (strcmp(refptr->segs[k].lm_s_src,"N")))  {
	/*printf("non SIL,N (%s)found at %d\n",refptr->segs[k].lm_s_src,k);*/
	num_real_matches_found++;
      }
      else  {
	/*printf("foo\n");*/
      }
    }
    else  {
      return(0);
    }
    if (num_real_matches_found == 2)  {
      *index_i = k;
      *index_j = l;
      return(1);
    }
    k++; 
    l++;
  }
  return(0);
}


/*******************************/
/*  Compares part of a ref force_out with a hyp force_out type
    starting from beginning of error region.  Finds end of error region.
    (May be the end of the utterance.)
*/

int find_rejoin0(refptr,hypptr,ress,hess,reesptr,heesptr)
force_out_t *refptr,*hypptr;
int32 ress,hess;
int32 *reesptr,*heesptr;

{
  int32 i,j;
  int match_found = 0;

  /* Goal of routine:  set *reesptr and *heesptr */
  /*printf("entered find_rejoin:  ress is %d, hess is %d\n",ress,hess);*/
  j = hess;
  for (i=ress;i<(refptr->num_segs)-1;i++)  {
    for (j=hess;j<(hypptr->num_segs)-1;j++)  {
      /*printf("checking %d, %d\n",i,j);
      fflush(stdout);*/
      /* why two ahead? changed 20 feb 96 */
#if 0
      if ((match_force_out_seg(&refptr->segs[i],&hypptr->segs[j])) &&
	  (match_force_out_seg(&refptr->segs[i+1],&hypptr->segs[j+1])))  {
#endif
      if ((match_force_out_seg(&refptr->segs[i],&hypptr->segs[j])))  {
	match_found = 1;
	/*printf("match found. i now %d j now %d\n",i,j);
	fflush(stdout);*/
	/* *heesptr = j;*/
	*heesptr = j-1;
	break;
      }
      /*printf("no match found.\n");*/
    }
    if (match_found)  {
      /* *reesptr = i; */
      *reesptr = i-1;
      break;
    }
  }
  if (match_found)  {
    /*printf("(match) rees found to be %d, hees is %d\n",*reesptr,*heesptr);
    fflush(stdout);*/
    return(match_found);
  }
  /* the end silences actually matched */
  else if (match_force_out_seg(&refptr->segs[(refptr->num_segs)-1],
			       &hypptr->segs[(hypptr->num_segs)-1]))  {
    *reesptr = (refptr->num_segs)-2;
    *heesptr = (hypptr->num_segs)-2;
    return(1);
  }
  /* the final error region went all the way to the end */
  else  {
    *reesptr = (refptr->num_segs)-1;
    *heesptr = (hypptr->num_segs)-1;
    /*printf("(no match) rees found to be %d, hees is %d\n",*reesptr,*heesptr);*/
    return(0);
  }
}

/*******************************/
/*  Compares part of a ref force_out with a hyp force_out type
    starting from beginning of error region.  Finds end of error region.
    (May be the end of the utterance.)

    BASED ON BIGRAM LANGUAGE MODEL.
*/

int find_rejoin1(refptr,hypptr,ress,hess,reesptr,heesptr)
force_out_t *refptr,*hypptr;
int32 ress,hess;
int32 *reesptr,*heesptr;

{
  int32 i,j;
  int match_found = 0;

  /* Goal of routine:  set *reesptr and *heesptr */
  /*printf("entered find_rejoin:  ress is %d, hess is %d\n",ress,hess);*/
  j = hess;
  for (i=ress;i<(refptr->num_segs)-1;i++)  {
    for (j=hess;j<(hypptr->num_segs)-1;j++)  {
      /*printf("checking %d, %d\n",i,j);*/
      if ((match_force_out_seg(&refptr->segs[i],&hypptr->segs[j])) &&
	  (match_force_out_seg(&refptr->segs[i+1],&hypptr->segs[j+1])))  {
	/* put this on a switch */
      /*if (end_of_error_region(refptr,hypptr,&i,&j))  {*/
	match_found = 1;
	/*printf("match found. i now %d j now %d\n",i,j);*/
	*heesptr = j;
	/* *heesptr = j-1;*/
	break;
      }
      /*printf("no match found.\n");*/
    }
    if (match_found)  {
      *reesptr = i; 
      /* *reesptr = i-1;*/
      break;
    }
  }
  if (match_found)  {
    /*printf("(match) rees found to be %d, hees is %d\n",*reesptr,*heesptr);*/
    return(match_found);
  }
  /* the end silences actually matched */
  /* else if (match_force_out_seg(&refptr->segs[(refptr->num_segs)-1], */
/* 			       &hypptr->segs[(hypptr->num_segs)-1]))  { */
/*     *reesptr = (refptr->num_segs)-2; */
/*     *heesptr = (hypptr->num_segs)-2; */
/*     return(1); */
/*   } */
  /* the final error region went all the way to the end */
  else  {
    *reesptr = (refptr->num_segs)-1;
    *heesptr = (hypptr->num_segs)-1;
    /*printf("(no match) rees found to be %d, hees is %d\n",*reesptr,*heesptr);*/
    return(0);
  }
}

/*******************************/
/*  Compares part of a ref force_out with a hyp force_out type
    starting from beginning of error region.  Finds end of error region.
    (May be the end of the utterance.)

    BASED ON TRIGRAM LANGUAGE MODEL.
*/

int find_rejoin2(refptr,hypptr,ress,hess,reesptr,heesptr)
force_out_t *refptr,*hypptr;
int32 ress,hess;
int32 *reesptr,*heesptr;

{
  int32 i,j;
  int match_found = 0;

  /* Goal of routine:  set *reesptr and *heesptr */
  /*printf("entered find_rejoin:  ress is %d, hess is %d\n",ress,hess);*/
  j = hess;
  for (i=ress;i<(refptr->num_segs)-1;i++)  {
    for (j=hess;j<(hypptr->num_segs)-1;j++)  {
      /*printf("checking %d, %d\n",i,j);*/
      if ((match_force_out_seg(&refptr->segs[i],&hypptr->segs[j])) &&
	  (match_force_out_seg(&refptr->segs[i+1],&hypptr->segs[j+1])))  {
	/* put this on a switch */
      /*if (end_of_error_region(refptr,hypptr,&i,&j))  {*/
	match_found = 1;
	/*printf("match found. i now %d j now %d\n",i,j);*/
	if ((j+1)<hypptr->num_segs)  {
	  *heesptr = j+1;
	}
	else  {
	  *heesptr = j;
	}
	/* *heesptr = j-1;*/
	break;
      }
      /*printf("no match found.\n");*/
    }
    if (match_found)  {
      if ((i+1)<refptr->num_segs)  {
	*reesptr = i+1;
      }
      else  {
	*reesptr = i;
      }
      /* *reesptr = i-1;*/
      break;
    }
  }
  if (match_found)  {
    /*printf("(match) rees found to be %d, hees is %d\n",*reesptr,*heesptr);
    fflush(stdout);*/
    return(match_found);
  }
  /* the end silences actually matched */
  /* else if (match_force_out_seg(&refptr->segs[(refptr->num_segs)-1], */
/* 			       &hypptr->segs[(hypptr->num_segs)-1]))  { */
/*     *reesptr = (refptr->num_segs)-2; */
/*     *heesptr = (hypptr->num_segs)-2; */
/*     return(1); */
/*   } */
  /* the final error region went all the way to the end */
  else  {
    *reesptr = (refptr->num_segs)-1;
    *heesptr = (hypptr->num_segs)-1;
    /*printf("(no match) rees found to be %d, hees is %d\n",*reesptr,*heesptr);*/
    return(0);
  }
}

/*******************************/
/*  Prints an error region to stdio:  Wayne's easy-to-read format */

void print0_err_region(refptr,hypptr,outfp,ress,rees,hess,hees,err_region,print_iter_num)
force_out_t *refptr,*hypptr;
FILE *outfp;
int32 ress,rees,hess,hees;
err_region_t *err_region;
int print_iter_num;
{

  int32 num_ref,num_hyp,num_segs;
  int i;
  int32 ref_tot_LM,hyp_tot_LM,ref_tot_AC,hyp_tot_AC;
  int32 ref_tot,hyp_tot;
  int num_groups,num_left,gr_size,this_group;


  num_ref = rees - ress + 1;

  num_hyp = hees - hess + 1;
  
  if (num_ref>=num_hyp)  {
    num_segs = num_ref;
  }
  else  {
    num_segs = num_hyp;
  }

  gr_size = 5;
  num_groups = (num_segs/gr_size)+1;
  num_left = num_segs%gr_size;
  /*  num_groups may be one too large if num_left is zero */
  if ((num_groups>1)&&(num_left==0))  {
    num_groups--;
    num_left=gr_size;
  }

  
  if (print_iter_num == 1)
    fprintf(outfp,"Error region:\n");
  else
    fprintf(outfp, "Error subregion:\n");

  this_group = 1;
  /* print all but the last group of segs */
  while (this_group<num_groups)  {
    fprintf(outfp,"REF:  ");
    /* print lm stuff for ref seg, i is offset from start seg */
    for (i=((this_group-1)*gr_size);i<(gr_size*this_group);i++)  {
      if (i<num_ref)  {
	fprintf(outfp," %15s ",refptr->segs[ress+i].word);
      }
      else  {
	fprintf(outfp,"                 ");
      }
    }
    fprintf(outfp,"\n");
    fprintf(outfp,"HYP:  ");
    /*  print lm stuff for hyp seg */
    for (i=((this_group-1)*gr_size);i<(gr_size*this_group);i++)  {
      if (i<num_hyp)  {
	fprintf(outfp," %15s ",hypptr->segs[hess+i].word);
      }
      else  {
	fprintf(outfp,"                 ");
      }
    }
    fprintf(outfp,"\n");
    fprintf(outfp,"\n");
    this_group++;
  }


  /* print last group of segs and totals (delta) information */
  fprintf(outfp,"      ");
  /* print headers */
  for (i=0;i<num_left;i++)  {
    fprintf(outfp,"                 ");
  }
  fprintf(outfp,"       LM  ");
  fprintf(outfp,"       AC  ");
  fprintf(outfp,"      TOT \n");

  fprintf(outfp,"REF:  ");
  /* print lm stuff for ref seg, i is offset from start seg */
  for (i=((this_group-1)*gr_size);i<num_segs;i++)  {
    if (i<num_ref)  {
      fprintf(outfp," %15s ",refptr->segs[ress+i].word);
    }
    else  {
      fprintf(outfp,"                 ");
    }
  }
  /* print score deltas for ref segs
     (remember better scores are less negative and thus bigger)*/
  if (err_region->ref_data.lm_s > err_region->hyp_data.lm_s)  {
    fprintf(outfp,"%11li",err_region->ref_data.lm_s - 
                          err_region->hyp_data.lm_s);
  }
  else  {
    fprintf(outfp,"           ");
  }
  if (err_region->ref_data.acoust_s > err_region->hyp_data.acoust_s)  {
    fprintf(outfp,"%11li",err_region->ref_data.acoust_s -
                          err_region->hyp_data.acoust_s);
  }
  else  {
    fprintf(outfp,"           ");
  }
  if (err_region->ref_data.tot_s > err_region->hyp_data.tot_s)  {
    ref_tot_bigger++;
    fprintf(outfp,"%11li\n", err_region->ref_data.tot_s -
                             err_region->hyp_data.tot_s);
  }
  else  {
    hyp_tot_bigger++;
    fprintf(outfp,"\n");
  }

  fprintf(outfp,"HYP:  ");
  /*  print lm stuff for hyp seg */
  for (i=((this_group-1)*gr_size);i<num_segs;i++)  {
    if (i<num_hyp)  {
      fprintf(outfp," %15s ",hypptr->segs[hess+i].word);
    }
    else  {
      fprintf(outfp,"                 ");
    }
  }
  /* print score deltas for hyp */
  if (err_region->ref_data.lm_s <= err_region->hyp_data.lm_s )  {
    fprintf(outfp,"%11li",err_region->hyp_data.lm_s -
                          err_region->ref_data.lm_s );
  }
  else  {
    fprintf(outfp,"           ");
  }
  if (err_region->ref_data.acoust_s <= err_region->hyp_data.acoust_s)  {
    fprintf(outfp,"%11li",err_region->hyp_data.acoust_s -
                          err_region->ref_data.acoust_s);
  }
  else  {
    fprintf(outfp,"           ");
  }
  if (err_region->ref_data.tot_s <= err_region->hyp_data.tot_s )  {
    fprintf(outfp,"%11li\n",err_region->hyp_data.tot_s -
                            err_region->ref_data.tot_s );
  }
  else  {
    fprintf(outfp,"\n");
  }


  /* Print out subregion information for all children and grandchildren, etc.
     of this error region. */

  if (err_region->num_children != 0)  {
    /*printf("num_children = %d \n", err_region->num_children);*/
    for(i = 1; i <= err_region->num_children; i++)  {
      print_iter_num += 1;
      /*printf("Calling print0_err_region for a subregion.\n");*/
      /*printf("i = %d\n",i);*/
      print0_err_region(refptr, hypptr, outfp,
                        err_region->child[i-1]->ref_data.st_seg,
                        err_region->child[i-1]->ref_data.en_seg,
                        err_region->child[i-1]->hyp_data.st_seg,
                        err_region->child[i-1]->hyp_data.en_seg,
                        err_region->child[i-1], print_iter_num);
    }
  }
}
/*******************************/
/*  Prints an error region to stdio:  Wayne's easy-to-read format 
    with individual scores */

void print1_err_region(refptr,hypptr,outfp,ress,rees,hess,hees,err_region,print_iter_num)
force_out_t *refptr,*hypptr;
FILE *outfp;
int32 ress,rees,hess,hees;
err_region_t *err_region;
int print_iter_num;

{
  int32 num_ref,num_hyp,num_segs;
  int i;
  int32 ref_tot_LM,hyp_tot_LM,ref_tot_AC,hyp_tot_AC;
  int32 ref_tot,hyp_tot;
  int num_groups,num_left,gr_size,this_group;

  num_ref = rees - ress + 1;
  num_hyp = hees - hess + 1;
  
  if (num_ref>=num_hyp)  {
    num_segs = num_ref;
  }
  else  {
    num_segs = num_hyp;
  }

  gr_size = 5;
  num_groups = (num_segs/gr_size)+1;
  num_left = num_segs%gr_size;
  /*  num_groups may be one too large if num_left is zero */
  if ((num_groups>1)&&(num_left==0))  {
    num_groups--;
    num_left=gr_size;
  }

  if (print_iter_num == 1)
    fprintf(outfp,"Error region:\n");
  else
    fprintf(outfp, "Error subregion:\n");

  this_group = 1;
  /* print all but the last group of segs */
  while (this_group<num_groups)  {
    fprintf(outfp,"REF:  ");
    /* print lm stuff for ref seg, i is offset from start seg */
    for (i=((this_group-1)*gr_size);i<(gr_size*this_group);i++)  {
      if (i<num_ref)  {
	fprintf(outfp," %11li %15s ",refptr->segs[ress+i].lm_s,refptr->segs[ress+i].word);
      }
      else  {
	fprintf(outfp,"                             ");
      }
    }
    fprintf(outfp,"\n");
    fprintf(outfp,"      ");
    /*  print acoustic stuff for ref seg */
    for (i=((this_group-1)*gr_size);i<(gr_size*this_group);i++)  {
      if (i<num_ref)  {
	fprintf(outfp,"     %23li ",refptr->segs[ress+i].acoust_s);
      }
      else  {
	fprintf(outfp,"                             ");
      }
    }
    fprintf(outfp,"\n");
    fprintf(outfp,"      ");
    /* print acoustic stuff for hyp seg */
    for (i=((this_group-1)*gr_size);i<(gr_size*this_group);i++)  {
      if (i<num_hyp)  {
	fprintf(outfp,"     %23li ",hypptr->segs[hess+i].acoust_s);
      }
      else  {
	fprintf(outfp,"                             ");
      }
    }
    fprintf(outfp,"\n");
    fprintf(outfp,"HYP:  ");
    /*  print lm stuff for hyp seg */
    for (i=((this_group-1)*gr_size);i<(gr_size*this_group);i++)  {
      if (i<num_hyp)  {
	fprintf(outfp," %11li %15s ",hypptr->segs[hess+i].lm_s,hypptr->segs[hess+i].word);
      }
      else  {
	fprintf(outfp,"                             ");
      }
    }
    fprintf(outfp,"\n");
    fprintf(outfp,"\n");
    this_group++;
  }

  /* print last group of segs and totals (delta) information */
  fprintf(outfp,"      ");
  /* print headers */
  for (i=0;i<num_left;i++)  {
    fprintf(outfp,"                             ");
  }
  fprintf(outfp,"       LM  ");
  fprintf(outfp,"       AC  ");
  fprintf(outfp,"      TOT \n");

  fprintf(outfp,"REF:  ");
  /* print lm stuff for ref seg, i is offset from start seg */
  for (i=((this_group-1)*gr_size);i<num_segs;i++)  {
    if (i<num_ref)  {
      fprintf(outfp," %11li %15s ",refptr->segs[ress+i].lm_s,refptr->segs[ress+i].word);
    }
    else  {
      fprintf(outfp,"                             ");
    }
  }
  /* print score deltas for ref segs
     (remember better scores are less negative and thus bigger)*/
  if (err_region->ref_data.lm_s > err_region->hyp_data.lm_s)  {
    fprintf(outfp,"%11li",err_region->ref_data.lm_s - 
                          err_region->hyp_data.lm_s);
  }
  else  {
    fprintf(outfp,"           ");
  }
  if (err_region->ref_data.acoust_s > err_region->hyp_data.acoust_s)  {
    fprintf(outfp,"%11li",err_region->ref_data.acoust_s -
                          err_region->hyp_data.acoust_s);
  }
  else  {
    fprintf(outfp,"           ");
  }
  if (err_region->ref_data.tot_s > err_region->hyp_data.tot_s)  {
    ref_tot_bigger++;
    fprintf(outfp,"%11li\n", err_region->ref_data.tot_s -
                             err_region->hyp_data.tot_s);
  }
  else  {
    hyp_tot_bigger++;
    fprintf(outfp,"\n");
  }
  fprintf(outfp,"      ");
  /*  print acoustic stuff for ref seg */
  for (i=((this_group-1)*gr_size);i<num_segs;i++)  {
    if (i<num_ref)  {
      fprintf(outfp,"     %23li ",refptr->segs[ress+i].acoust_s);
    }
    else  {
      fprintf(outfp,"                             ");
    }
  }
  fprintf(outfp,"\n");
  fprintf(outfp,"      ");
  /* print acoustic stuff for hyp seg */
  for (i=((this_group-1)*gr_size);i<num_segs;i++)  {
    if (i<num_hyp)  {
      fprintf(outfp,"     %23li ",hypptr->segs[hess+i].acoust_s);
    }
    else  {
      fprintf(outfp,"                             ");
    }
  }
  fprintf(outfp,"\n");
  fprintf(outfp,"HYP:  ");
  /*  print lm stuff for hyp seg */
  for (i=((this_group-1)*gr_size);i<num_segs;i++)  {
    if (i<num_hyp)  {
      fprintf(outfp," %11li %15s ",hypptr->segs[hess+i].lm_s,hypptr->segs[hess+i].word);
    }
    else  {
      fprintf(outfp,"                             ");
    }
  }
  /* print score deltas for hyp */
  if (err_region->ref_data.lm_s <= err_region->hyp_data.lm_s)  {
    fprintf(outfp,"%11li",err_region->hyp_data.lm_s -
                          err_region->ref_data.lm_s );
  }
  else  {
    fprintf(outfp,"           ");
  }
  if (err_region->ref_data.acoust_s <= err_region->hyp_data.acoust_s)  {
    fprintf(outfp,"%11li",err_region->hyp_data.acoust_s -
                          err_region->ref_data.acoust_s);
  }
  else  {
    fprintf(outfp,"           ");
  }
  if (err_region->ref_data.tot_s <= err_region->hyp_data.tot_s)  {
    fprintf(outfp,"%11li\n",err_region->hyp_data.tot_s -
                            err_region->ref_data.tot_s);
  }
  else  {
    fprintf(outfp,"\n");
  }

  /* Print out subregion information for all children and grandchildren, etc.
     of this error region. */

  if (err_region->num_children != 0)  {
    /*printf("num_children = %d \n", err_region->num_children);*/
    for(i = 1; i <= err_region->num_children; i++)  {
      print_iter_num += 1;
      /*printf("Calling print1_err_region for a subregion.\n");*/
      /*printf("i = %d\n",i);*/
      print1_err_region(refptr, hypptr, outfp,
                        err_region->child[i-1]->ref_data.st_seg,
                        err_region->child[i-1]->ref_data.en_seg,
                        err_region->child[i-1]->hyp_data.st_seg,
                        err_region->child[i-1]->hyp_data.en_seg,
                        err_region->child[i-1], print_iter_num);
    }
  }
}

/*******************************/
/*  Prints an error region to stdio:  Wayne's plus frames format */

void print2_err_region(refptr,hypptr,outfp,ress,rees,hess,hees,err_region,print_iter_num)
force_out_t *refptr,*hypptr;
FILE *outfp;
int32 ress,rees,hess,hees;
err_region_t *err_region;
int print_iter_num;
{
  int32 num_ref,num_hyp,num_segs;
  int i;
  int32 ref_tot_LM,hyp_tot_LM,ref_tot_AC,hyp_tot_AC;
  int32 ref_tot,hyp_tot;
  int num_groups,num_left,gr_size,this_group;

  num_ref = rees - ress + 1;
  num_hyp = hees - hess + 1;
  
  if (num_ref>=num_hyp)  {
    num_segs = num_ref;
  }
  else  {
    num_segs = num_hyp;
  }

  gr_size = 5;
  num_groups = (num_segs/gr_size)+1;
  num_left = num_segs%gr_size;
  /*  num_groups may be one too large if num_left is zero */
  if ((num_groups>1)&&(num_left==0))  {
    num_groups--;
    num_left=gr_size;
  }


  if (print_iter_num == 1)
    fprintf(outfp,"Error region:\n");
  else
    fprintf(outfp, "Error subregion:\n");

  this_group = 1;
  /* print all but the last group of segs */
  while (this_group<num_groups)  {
    /*printf("2");*/
    /*  print frames for ref seg */
    fprintf(outfp,"      ");
    for (i=((this_group-1)*gr_size);i<(gr_size*this_group);i++)  {
      if (i<num_ref)  {
	fprintf(outfp,"               (%5d,%5d) ",
	       refptr->segs[ress+i].start_f,refptr->segs[ress+i].end_f);
      }
      else  {
	fprintf(outfp,"                             ");
      }
    }
    fprintf(outfp,"\n");
    fprintf(outfp,"REF:  ");
    /* print lm stuff for ref seg, i is offset from start seg */
    for (i=((this_group-1)*gr_size);i<(gr_size*this_group);i++)  {
      if (i<num_ref)  {
	fprintf(outfp," %11li %15s ",refptr->segs[ress+i].lm_s,refptr->segs[ress+i].word);
      }
      else  {
	fprintf(outfp,"                             ");
      }
    }
    fprintf(outfp,"\n");
    fprintf(outfp,"      ");
    /*  print acoustic stuff for ref seg */
    for (i=((this_group-1)*gr_size);i<(gr_size*this_group);i++)  {
      if (i<num_ref)  {
	fprintf(outfp,"     %23li ",refptr->segs[ress+i].acoust_s);
      }
      else  {
	fprintf(outfp,"                             ");
      }
    }
    fprintf(outfp,"\n");
    fprintf(outfp,"      ");
    /* print acoustic stuff for hyp seg */
    for (i=((this_group-1)*gr_size);i<(gr_size*this_group);i++)  {
      if (i<num_hyp)  {
	fprintf(outfp,"     %23li ",hypptr->segs[hess+i].acoust_s);
      }
      else  {
	fprintf(outfp,"                             ");
      }
    }
    fprintf(outfp,"\n");
    fprintf(outfp,"HYP:  ");
    /*  print lm stuff for hyp seg */
    for (i=((this_group-1)*gr_size);i<(gr_size*this_group);i++)  {
      if (i<num_hyp)  {
	fprintf(outfp," %11li %15s ",hypptr->segs[hess+i].lm_s,hypptr->segs[hess+i].word);
      }
      else  {
	fprintf(outfp,"                             ");
      }
    }
    fprintf(outfp,"\n");
    fprintf(outfp,"      ");
    /*  print frames for hyp seg */
    for (i=((this_group-1)*gr_size);i<(gr_size*this_group);i++)  {
      if (i<num_hyp)  {
	fprintf(outfp,"               (%5d,%5d) ",
	       hypptr->segs[hess+i].start_f,hypptr->segs[hess+i].end_f);
      }
      else  {
	fprintf(outfp,"                             ");
      }
    }
    fprintf(outfp,"\n");
    fprintf(outfp,"\n");
    this_group++;
  }

  /* print last group of segs and totals (delta) information */
  fprintf(outfp,"      ");
  /* print headers */
  for (i=0;i<num_left;i++)  {
    fprintf(outfp,"                             ");
  }
  fprintf(outfp,"       LM  ");
  fprintf(outfp,"       AC  ");
  fprintf(outfp,"      TOT \n");

  fprintf(outfp,"      ");
  for (i=((this_group-1)*gr_size);i<num_segs;i++)  {
    if (i<num_ref)  {
      fprintf(outfp,"               (%5d,%5d) ",
	     refptr->segs[ress+i].start_f,refptr->segs[ress+i].end_f);
    }
    else  {
      fprintf(outfp,"                             ");
    }
  }
  fprintf(outfp,"\n");
  fprintf(outfp,"REF:  ");
  /* print lm stuff for ref seg, i is offset from start seg */
  for (i=((this_group-1)*gr_size);i<num_segs;i++)  {
    if (i<num_ref)  {
      fprintf(outfp," %11li %15s ",refptr->segs[ress+i].lm_s,refptr->segs[ress+i].word);
    }
    else  {
      fprintf(outfp,"                             ");
    }
  }
  /* print score deltas for ref segs
     (remember better scores are less negative and thus bigger)*/
  if (err_region->ref_data.lm_s > err_region->hyp_data.lm_s)  {
    fprintf(outfp,"%11li",err_region->ref_data.lm_s - 
                          err_region->hyp_data.lm_s);
  }
  else  {
    fprintf(outfp,"           ");
  }
  if (err_region->ref_data.acoust_s > err_region->hyp_data.acoust_s)  {
    fprintf(outfp,"%11li",err_region->ref_data.acoust_s -
                          err_region->hyp_data.acoust_s);
  }
  else  {
    fprintf(outfp,"           ");
  }
  if (err_region->ref_data.tot_s > err_region->hyp_data.tot_s)  {
    ref_tot_bigger++;
    fprintf(outfp,"%11li\n",err_region->ref_data.tot_s -
                            err_region->hyp_data.tot_s);
  }
  else  {
    hyp_tot_bigger++;
    fprintf(outfp,"\n");
  }
  fprintf(outfp,"      ");
  /*  print acoustic stuff for ref seg */
  for (i=((this_group-1)*gr_size);i<num_segs;i++)  {
    if (i<num_ref)  {
      fprintf(outfp,"     %23li ",refptr->segs[ress+i].acoust_s);
    }
    else  {
      fprintf(outfp,"                             ");
    }
  }
  fprintf(outfp,"\n");
  fprintf(outfp,"      ");
  /* print acoustic stuff for hyp seg */
  for (i=((this_group-1)*gr_size);i<num_segs;i++)  {
    if (i<num_hyp)  {
      fprintf(outfp,"     %23li ",hypptr->segs[hess+i].acoust_s);
    }
    else  {
      fprintf(outfp,"                             ");
    }
  }
  fprintf(outfp,"\n");
  fprintf(outfp,"HYP:  ");
  /*  print lm stuff for hyp seg */
  for (i=((this_group-1)*gr_size);i<num_segs;i++)  {
    if (i<num_hyp)  {
      fprintf(outfp," %11li %15s ",hypptr->segs[hess+i].lm_s,hypptr->segs[hess+i].word);
    }
    else  {
      fprintf(outfp,"                             ");
    }
  }
  /* print score deltas for hyp */
  if (err_region->ref_data.lm_s <= err_region->hyp_data.lm_s )  {
    fprintf(outfp,"%11li",err_region->hyp_data.lm_s -
                          err_region->ref_data.lm_s);
  }
  else  {
    fprintf(outfp,"           ");
  }
  if (err_region->ref_data.acoust_s <= err_region->hyp_data.acoust_s)  {
    fprintf(outfp,"%11li",err_region->hyp_data.acoust_s -
                          err_region->ref_data.acoust_s);
  }
  else  {
    fprintf(outfp,"           ");
  }
  if (err_region->ref_data.tot_s <= err_region->hyp_data.tot_s)  {
    fprintf(outfp,"%11li\n",err_region->hyp_data.tot_s -
                            err_region->ref_data.tot_s);
  }
  else  {
    fprintf(outfp,"\n");
  }
  fprintf(outfp,"      ");
  /*  print frames for hyp seg */
  for (i=((this_group-1)*gr_size);i<(gr_size*this_group);i++)  {
    if (i<num_hyp)  {
      fprintf(outfp,"               (%5d,%5d) ",
	     hypptr->segs[hess+i].start_f,hypptr->segs[hess+i].end_f);
    }
    else  {
      fprintf(outfp,"                             ");
    }
    fprintf(outfp,"\n");
  }  

  /* Print out subregion information for all children and grandchildren, etc.
     of this error region. */

  if (err_region->num_children != 0)  {
    /*printf("num_children = %d \n", err_region->num_children);*/
    for(i = 1; i <= err_region->num_children; i++)  {
      print_iter_num += 1;
      /*printf("Calling print2_err_region for a subregion.\n");*/
      /*printf("i = %d\n",i);*/
      print2_err_region(refptr, hypptr, outfp,
                        err_region->child[i-1]->ref_data.st_seg,
                        err_region->child[i-1]->ref_data.en_seg,
                        err_region->child[i-1]->hyp_data.st_seg,
                        err_region->child[i-1]->hyp_data.en_seg,
                        err_region->child[i-1], print_iter_num);
    }
  }
}


/*******************************/
/*  Prints an error region to stdio:  Wayne's plus frames and LM source information */

void print3_err_region (refptr, hypptr, outfp, ress, rees, hess, hees, err_region, print_iter_num)
force_out_t *refptr,*hypptr;
FILE *outfp;
int32 ress,rees,hess,hees;
err_region_t *err_region;
int print_iter_num;
{
  int32 num_ref,num_hyp,num_segs;
  int i;
  int32 ref_tot_LM,hyp_tot_LM,ref_tot_AC,hyp_tot_AC;
  int32 ref_tot,hyp_tot;
  int num_groups,num_left,gr_size,this_group;

  num_ref = rees - ress + 1;
  num_hyp = hees - hess + 1;
  
  if (num_ref>=num_hyp)  {
    num_segs = num_ref;
  }
  else  {
    num_segs = num_hyp;
  }

  gr_size = 5;
  num_groups = (num_segs/gr_size)+1;
  num_left = num_segs%gr_size;
  /*  num_groups may be one too large if num_left is zero */
  if ((num_groups>1)&&(num_left==0))  {
    num_groups--;
    num_left=gr_size;
  }


  if (print_iter_num == 1)
    fprintf(outfp,"Error region:\n");
  else
    fprintf(outfp, "Error subregion:\n");

  this_group = 1;
  /* print all but the last group of segs */
  while (this_group<num_groups)  {
    /*  print LM score source and frames for ref seg */
    fprintf(outfp,"      ");
    for (i=((this_group-1)*gr_size);i<(gr_size*this_group);i++)  {
      if (i<num_ref)  {
	fprintf(outfp,"       %5s - (%5d,%5d) ",refptr->segs[ress+i].lm_s_src,
	       refptr->segs[ress+i].start_f,refptr->segs[ress+i].end_f);
      }
      else  {
	fprintf(outfp,"                             ");
      }
    }
    fprintf(outfp,"\n");
    fprintf(outfp,"REF:  ");
    /* print lm stuff for ref seg, i is offset from start seg */
    for (i=((this_group-1)*gr_size);i<(gr_size*this_group);i++)  {
      if (i<num_ref)  {
	fprintf(outfp," %11li %15s ",refptr->segs[ress+i].lm_s,refptr->segs[ress+i].word);
      }
      else  {
	fprintf(outfp,"                             ");
      }
    }
    fprintf(outfp,"\n");
    fprintf(outfp,"      ");
    /*  print acoustic stuff for ref seg */
    for (i=((this_group-1)*gr_size);i<(gr_size*this_group);i++)  {
      if (i<num_ref)  {
	fprintf(outfp,"     %23li ",refptr->segs[ress+i].acoust_s);
      }
      else  {
	fprintf(outfp,"                             ");
      }
    }
    fprintf(outfp,"\n");
    fprintf(outfp,"      ");
    /* print acoustic stuff for hyp seg */
    for (i=((this_group-1)*gr_size);i<(gr_size*this_group);i++)  {
      if (i<num_hyp)  {
	fprintf(outfp,"     %23li ",hypptr->segs[hess+i].acoust_s);
      }
      else  {
	fprintf(outfp,"                             ");
      }
    }
    fprintf(outfp,"\n");
    fprintf(outfp,"HYP:  ");
    /*  print lm stuff for hyp seg */
    for (i=((this_group-1)*gr_size);i<(gr_size*this_group);i++)  {
      if (i<num_hyp)  {
	fprintf(outfp," %11li %15s ",hypptr->segs[hess+i].lm_s,hypptr->segs[hess+i].word);
      }
      else  {
	fprintf(outfp,"                             ");
      }
    }
    fprintf(outfp,"\n");
    fprintf(outfp,"      ");
    /*  print LM Score source and frames for hyp seg */
    for (i=((this_group-1)*gr_size);i<(gr_size*this_group);i++)  {
      if (i<num_hyp)  {
	fprintf(outfp,"       %5s - (%5d,%5d) ",hypptr->segs[hess+i].lm_s_src,
	       hypptr->segs[hess+i].start_f,hypptr->segs[hess+i].end_f);
      }
      else  {
	fprintf(outfp,"                             ");
      }
    }
    fprintf(outfp,"\n");
    fprintf(outfp,"\n");
    this_group++;
  }

  /* print last group of segs and totals (delta) information */
  fprintf(outfp,"      ");
  /* print headers */
  for (i=0;i<num_left;i++)  {
    fprintf(outfp,"                             ");
  }
  fprintf(outfp,"       LM  ");
  fprintf(outfp,"       AC  ");
  fprintf(outfp,"      TOT \n");

  fprintf(outfp,"      ");
  /* print LM score source and frames for ref segs */
  for (i=((this_group-1)*gr_size);i<num_segs;i++)  {
    if (i<num_ref)  {
      fprintf(outfp,"       %5s - (%5d,%5d) ",refptr->segs[ress+i].lm_s_src,
	     refptr->segs[ress+i].start_f,refptr->segs[ress+i].end_f);
    }
    else  {
      fprintf(outfp,"                             ");
    }
  }
  fprintf(outfp,"\n");
  fprintf(outfp,"REF:  ");
  /* print lm stuff for ref seg, i is offset from start seg */
  for (i=((this_group-1)*gr_size);i<num_segs;i++)  {
    if (i<num_ref)  {
      fprintf(outfp," %11li %15s ",refptr->segs[ress+i].lm_s,refptr->segs[ress+i].word);
    }
    else  {
      fprintf(outfp,"                             ");
    }
  }
  /* print score deltas for ref segs
     (remember better scores are less negative and thus bigger)*/
  if (err_region->ref_data.lm_s > err_region->hyp_data.lm_s)  {
    fprintf(outfp,"%11li",err_region->ref_data.lm_s - 
                          err_region->hyp_data.lm_s);
  }
  else  {
    fprintf(outfp,"           ");
  }
  if (err_region->ref_data.acoust_s > err_region->hyp_data.acoust_s)  {
    fprintf(outfp,"%11li",err_region->ref_data.acoust_s -
                          err_region->hyp_data.acoust_s);
  }
  else  {
    fprintf(outfp,"           ");
  }
  if (err_region->ref_data.tot_s > err_region->hyp_data.tot_s)  {
    fprintf(outfp,"%11li\n",err_region->ref_data.tot_s -
                            err_region->hyp_data.tot_s);
  }
  else  {
    fprintf(outfp,"\n");
  }
  fprintf(outfp,"      ");
  /*  print acoustic stuff for ref seg */
  for (i=((this_group-1)*gr_size);i<num_segs;i++)  {
    if (i<num_ref)  {
      fprintf(outfp,"     %23li ",refptr->segs[ress+i].acoust_s);
    }
    else  {
      fprintf(outfp,"                             ");
    }
  }
  fprintf(outfp,"\n");
  fprintf(outfp,"      ");
  /* print acoustic stuff for hyp seg */
  for (i=((this_group-1)*gr_size);i<num_segs;i++)  {
    if (i<num_hyp)  {
      fprintf(outfp,"     %23li ",hypptr->segs[hess+i].acoust_s);
    }
    else  {
      fprintf(outfp,"                             ");
    }
  }
  fprintf(outfp,"\n");
  fprintf(outfp,"HYP:  ");
  /*  print lm stuff for hyp seg */
  for (i=((this_group-1)*gr_size);i<num_segs;i++)  {
    if (i<num_hyp)  {
      fprintf(outfp," %11li %15s ",hypptr->segs[hess+i].lm_s,hypptr->segs[hess+i].word);
    }
    else  {
      fprintf(outfp,"                             ");
    }
  }
  /* print score deltas for hyp */
  if (err_region->ref_data.lm_s <= err_region->hyp_data.lm_s)  {
    fprintf(outfp,"%11li",err_region->hyp_data.lm_s -
                          err_region->ref_data.lm_s);
  }
  else  {
    fprintf(outfp,"           ");
  }
  if (err_region->ref_data.acoust_s <= err_region->hyp_data.acoust_s)  {
    fprintf(outfp,"%11li",err_region->hyp_data.acoust_s -
                          err_region->ref_data.acoust_s);
  }
  else  {
    fprintf(outfp,"           ");
  }
  if (err_region->ref_data.tot_s <= err_region->hyp_data.tot_s)  {
    fprintf(outfp,"%11li\n",err_region->hyp_data.tot_s -
                            err_region->ref_data.tot_s);
  }
  else  {
    fprintf(outfp,"\n");
  }
  fprintf(outfp,"      ");
  /*  print LM Score source and frames for hyp seg */
  for (i=((this_group-1)*gr_size);i<(gr_size*this_group);i++)  {
    if (i<num_hyp)  {
      fprintf(outfp,"       %5s - (%5d,%5d) ",hypptr->segs[hess+i].lm_s_src,
	     hypptr->segs[hess+i].start_f,hypptr->segs[hess+i].end_f);
    }
    else  {
      fprintf(outfp,"                             ");
    }
  }
  fprintf(outfp,"\n");
  /* pruning error region */
  if (err_region->ref_data.tot_s > err_region->hyp_data.tot_s)  {
    if ((err_region->ref_data.lm_s > err_region->hyp_data.lm_s) &&
	(err_region->ref_data.acoust_s > err_region->hyp_data.acoust_s))  {
      prune_regs_both++;
      prune_both_ratio_lmtoa += (float) ((err_region->ref_data.lm_s - err_region->hyp_data.lm_s)/(err_region->ref_data.acoust_s - err_region->hyp_data.acoust_s));
    }
    else if (err_region->ref_data.lm_s > err_region->hyp_data.lm_s)  {
      prune_regs_lm++;
    }
    else  {
      prune_regs_ac++;
    }
  }
  /* nonpruning error */
  else  {
    if ((err_region->hyp_data.lm_s > err_region->ref_data.lm_s) &&
	(err_region->hyp_data.acoust_s > err_region->ref_data.acoust_s))  {
      no_prune_regs_both++;
      no_prune_both_ratio_lmtoa += (float) ((err_region->hyp_data.lm_s - err_region->ref_data.lm_s)/(err_region->hyp_data.acoust_s - err_region->ref_data.acoust_s));
    }
    else if (err_region->hyp_data.lm_s > err_region->ref_data.lm_s)  {
      no_prune_regs_lm++;
    }
    else  {
      no_prune_regs_ac++;
    }
  }


  /* Print out subregion information for all children and grandchildren, etc.
     of this error region. */

  if (err_region->num_children != 0)  {
    /*printf("num_children = %d \n", err_region->num_children);*/
    for(i = 1; i <= err_region->num_children; i++)  {
      print_iter_num += 1;
      /*printf("Calling print3_err_region for a subregion.\n");*/
      /*printf("i = %d\n",i);*/
      print3_err_region(refptr, hypptr, outfp,
                        err_region->child[i-1]->ref_data.st_seg,
                        err_region->child[i-1]->ref_data.en_seg,
                        err_region->child[i-1]->hyp_data.st_seg,
                        err_region->child[i-1]->hyp_data.en_seg,
                        err_region->child[i-1], print_iter_num);
    }
  }
}

int notSILs(err_reg,ress,rees,hess,hees,refptr,hypptr)
err_region_t *err_reg;
int32 ress,rees,hess,hees;
force_out_t *refptr,*hypptr;
{
  int i;
  int notSILs = 0;

  for (i=ress;i<=rees;i++)  {
    if (!((!strcmp(refptr->segs[i].word,"SIL")) ||
	  (!strcmp(refptr->segs[i].word,"<s>")) ||
	  (!strcmp(refptr->segs[i].word,"</s>"))))  {
      notSILs = 1;
      break;
    }
  }

  for (i=hess;i<=hees;i++)  {
    if (!((!strcmp(hypptr->segs[i].word,"SIL")) ||
	  (!strcmp(hypptr->segs[i].word,"<s>")) ||
	  (!strcmp(hypptr->segs[i].word,"</s>"))))  {
      notSILs = 1;
      break;
    }
  }

  if (!notSILs)  {
    /*printf("Skipping error region because it contains only SILs.\n");*/
  }
  return(notSILs);
}

/****************************/
int ref_phone_found(refptr,ress,rees,hyp_phone,hyp_sf,hyp_ef)
force_out_t *refptr;
int32 ress,rees;
char *hyp_phone;
int hyp_sf,hyp_ef;

{
  int i,k;
  int32 num_ref;
  int num_ref_phones;

  num_ref = rees - ress + 1;

  /*printf("Looking for hyp phone match %s(%d,%d)\n",
	 hyp_phone,hyp_sf,hyp_ef);
  fflush(stdout);
  */

  for (i=0;i<num_ref;i++)  {
    num_ref_phones = refptr->segs[ress+i].phone_seq.num_phones;
    for (k=0;k<num_ref_phones;k++)  {
      /*printf("Checking ref phone %s(%d,%d)\n",
	     refptr->segs[ress+i].phone_seq.phones[k].name,
	     refptr->segs[ress+i].phone_seq.phones[k].sf,
	     refptr->segs[ress+i].phone_seq.phones[k].ef);
      fflush(stdout);
      */
      /* this is not the right phone, stop looking and go to next phone */
      if (strcmp(refptr->segs[ress+i].phone_seq.phones[k].name,hyp_phone))  {
	continue;
      }
      /* right phone, does its segmentation match within tolerance? */
      else if ((abs(refptr->segs[ress+i].phone_seq.phones[k].sf - hyp_sf)
		<= frame_tol) &&
	       (abs(refptr->segs[ress+i].phone_seq.phones[k].ef - hyp_ef)
		<= frame_tol))  {
	/* bingo */
	return(1);
      }
    }
  }
  return(0);
}


/****************************/
/* decide which HYP phones in this error region can be considered as
   correct and which cannot */
void tag_phones(refptr,hypptr,ress,rees,hess,hees)
force_out_t *refptr,*hypptr;
int32 ress,rees,hess,hees;
{
  int i,j;
  int32 num_ref,num_hyp;
  int num_ref_phones,num_hyp_phones;
  char hyp_phone[64];
  int hyp_sf,hyp_ef;

  hyp_phone[0] = '\0';

  num_ref = rees - ress + 1;
  num_hyp = hees - hess + 1;  

  /* the global variable that holds frame tolerance is frame_tol */

  /* for each ref phone, mark it as being in an error region */
  /*  for now, all ref phones will maintain their default correct label */
  for (i=0;i<num_ref;i++)  {
    num_ref_phones = refptr->segs[ress+i].phone_seq.num_phones;
    for(j=0;j<num_ref_phones;j++)  {
      refptr->segs[ress+i].phone_seq.phones[j].err_reg = 1;
    }
  }

  /* for each hyp phone, mark it as being in an error region and then
     try to find a ref phone that makes it
     seem correct (both are tagged as correct by default).  
     if not found, tag hyp phone as incorrect */
  for (i=0;i<num_hyp;i++)  {
    num_hyp_phones = hypptr->segs[hess+i].phone_seq.num_phones;
    for (j=0;j<num_hyp_phones;j++)  {
      hypptr->segs[hess+i].phone_seq.phones[j].err_reg=1;
      strcpy(hyp_phone,hypptr->segs[hess+i].phone_seq.phones[j].name);
      hyp_sf = hypptr->segs[hess+i].phone_seq.phones[j].sf;
      hyp_ef = hypptr->segs[hess+i].phone_seq.phones[j].ef;
      if (ref_phone_found(refptr,ress,rees,hyp_phone,hyp_sf,hyp_ef))  {
	/*printf("Phone %s(%d,%d) looks OK.\n",hyp_phone,hyp_sf,hyp_ef);
	fflush(stdout);*/
      }
      else  {
	hypptr->segs[hess+i].phone_seq.phones[j].correct = 0;
	/*printf("Phone %s(%d,%d) is incorrect.\n",hyp_phone,hyp_sf,hyp_ef);
	fflush(stdout);*/
      }
    }
  }
}
    

/*******************************/
void print_visual_err_region (refptr, hypptr, uttid, tkdir, ress, rees, hess, hees, err_region, print_iter_num, reg_no, norm, sc_ratio)
force_out_t *refptr,*hypptr;
char *uttid,*tkdir;
int32 ress,rees,hess,hees;
err_region_t *err_region;
int print_iter_num;
int reg_no;
int norm;
int sc_ratio;

{
  char newfile[256];
  int32 num_ref,num_hyp;
  int i,j;
  int32 ac_tot,lm_tot,tot;
  int ref_ac, ref_lm, ref_tot = 0;
  FILE *tkfp;
  int32 num_frames;
  int num_phones;
  int32 num_ph_frames;
  int unk_reg = 0;

  sprintf(newfile,"%s/%s.tk",tkdir,uttid);
  tkfp = CM_fopen(newfile,"a");

  /* decide which HYP phones in this error region can be considered as
     correct and which cannot */
  tag_phones(refptr,hypptr,ress,rees,hess,hees);

  num_ref = rees - ress + 1;
  num_hyp = hees - hess + 1;

  /*  will handle subregions later */

  /* set AC variables */
  if (err_region->ref_data.acoust_s > err_region->hyp_data.acoust_s)  {
    ref_ac = 1;
    ac_tot = err_region->ref_data.acoust_s - err_region->hyp_data.acoust_s;
  }
  else  {
    ref_ac = 0;
    ac_tot = err_region->hyp_data.acoust_s - err_region->ref_data.acoust_s;
  }
  /* set LM variables */
  if (err_region->ref_data.lm_s > err_region->hyp_data.lm_s)  {
    ref_lm= 1;
    lm_tot = err_region->ref_data.lm_s - err_region->hyp_data.lm_s;
  }
  else  {
    ref_lm = 0;
    lm_tot = err_region->hyp_data.lm_s - err_region->ref_data.lm_s;
  }
  /* set tot variables */
  if (err_region->ref_data.tot_s > err_region->hyp_data.tot_s)  {
    ref_tot= 1;
    tot = err_region->ref_data.tot_s - err_region->hyp_data.tot_s;
  }
  else  {
    ref_tot = 0;
    tot = err_region->hyp_data.tot_s - err_region->ref_data.tot_s;
  }

  /*printf("sc_ratio is %d\n",sc_ratio);
  fflush(stdout);*/

  /* write out general information */
  fprintf(tkfp,"tk_errwin_print_err %d %d %d %d %d %d %d %d %d\n",
	  refptr->segs[ress].start_f,	
	  refptr->segs[rees].end_f,
	  ref_lm,
	  (lm_tot / ((int32) sc_ratio)),
	  ref_ac,
	  (ac_tot / ((int32) sc_ratio)),
	  ref_tot,
	  (tot / ((int32) sc_ratio)),
	  reg_no);  

  /* write out REF information */
  for (i=0;i<num_ref;i++)  {
    num_phones = refptr->segs[ress+i].phone_seq.num_phones;
    if (strcmp(refptr->segs[ress+i].lm_s_src,"UNK")==0)  {
      unk_reg = 1;
    }
    if (norm == 1)  {
      num_frames = (refptr->segs[ress+i].end_f -
		    refptr->segs[ress+i].start_f);
      if (num_frames > 0)  {
	fprintf(tkfp,"tk_errwin_print_ref_word %d %d %s %s %d %d\n",
		refptr->segs[ress+i].start_f,
		refptr->segs[ress+i].end_f,
		refptr->segs[ress+i].word,
		refptr->segs[ress+i].lm_s_src,
		(refptr->segs[ress+i].lm_s / ((int32) sc_ratio)),
		(((int32)(refptr->segs[ress+i].acoust_s / num_frames)) /
		 ((int32) sc_ratio)));
      }
      else  {
	fprintf(tkfp,"tk_errwin_print_ref_word %d %d %s %s %d %d\n",
		refptr->segs[ress+i].start_f,
		refptr->segs[ress+i].end_f,
		refptr->segs[ress+i].word,
		refptr->segs[ress+i].lm_s_src,
		(refptr->segs[ress+i].lm_s / ((int32) sc_ratio)),
		0);
      }
      for (j=0;j<num_phones;j++)  {
	num_ph_frames = (int32) (refptr->segs[ress+i].phone_seq.phones[j].ef -
				 refptr->segs[ress+i].phone_seq.phones[j].sf);
	if (num_ph_frames > 0)  {
	  fprintf(tkfp,"tk_errwin_print_ref_phone %d %d %s %d\n",
		  refptr->segs[ress+i].phone_seq.phones[j].sf,
		  refptr->segs[ress+i].phone_seq.phones[j].ef,
		  refptr->segs[ress+i].phone_seq.phones[j].basename,
		  (((int32) (refptr->segs[ress+i].phone_seq.phones[j].ac_score /
			   num_ph_frames)) /
		   ((int32) sc_ratio)));
	}
	else  {
	  fprintf(tkfp,"tk_errwin_print_ref_phone %d %d %s %d\n",
		  refptr->segs[ress+i].phone_seq.phones[j].sf,
		  refptr->segs[ress+i].phone_seq.phones[j].ef,
		  refptr->segs[ress+i].phone_seq.phones[j].basename,
		  0);
	}
      }
    }
    else  {
      fprintf(tkfp,"tk_errwin_print_ref_word %d %d %s %s %d %d\n",
	      refptr->segs[ress+i].start_f,
	      refptr->segs[ress+i].end_f,
	      refptr->segs[ress+i].word,
	      refptr->segs[ress+i].lm_s_src,
	      (refptr->segs[ress+i].lm_s / ((int32) sc_ratio)),
	      (refptr->segs[ress+i].acoust_s / ((int32) sc_ratio)));
      for (j=0;j<num_phones;j++)  {
	fprintf(tkfp,"tk_errwin_print_ref_phone %d %d %s %d\n",
		refptr->segs[ress+i].phone_seq.phones[j].sf,
		refptr->segs[ress+i].phone_seq.phones[j].ef,
		refptr->segs[ress+i].phone_seq.phones[j].basename,
		(refptr->segs[ress+i].phone_seq.phones[j].ac_score / 
		 ((int32) sc_ratio)));
      }
    }
  }

  /* set various global counter variables by looking at ref_ac, ref_lm, and ref_tot*/
  /* was this error region due to an OOV word? */
  if (unk_reg == 1)  {
    num_unk_error_regions++;
  }
  else  {
    /* pruning error */
    if (ref_tot)  {
      ref_tot_bigger++;
      if (ref_ac)  {
	/* due to combined effect */
	if (ref_lm)  {
	  prune_regs_both++;
	}
	/* due to AC only */
	else  {
	  prune_regs_ac++;}
      }
      /* due to LM only */
      else  {
	prune_regs_lm++;
      }
    }
    /* modeling error */
    else  {
      hyp_tot_bigger++;
      if (!ref_ac)  {
	/* due to combined effect */
	if (!ref_lm)  {
	  no_prune_regs_both++;
	}
	/* due to AC only */
	else  {
	  no_prune_regs_ac++;
	}
      }
      /* due to LM only */
      else  {
	no_prune_regs_lm++;
      }
    }
  }

  /* write out HYP information */
  for (i=0;i<num_hyp;i++)  {
    num_phones = hypptr->segs[hess+i].phone_seq.num_phones;
    if (norm == 1)  {
      num_frames = (hypptr->segs[hess+i].end_f -
		    hypptr->segs[hess+i].start_f);
      if (num_frames > 0)  {
	fprintf(tkfp,"tk_errwin_print_hyp_word %d %d %s %s %d %d\n",
		hypptr->segs[hess+i].start_f,
		hypptr->segs[hess+i].end_f,
		hypptr->segs[hess+i].word,
		hypptr->segs[hess+i].lm_s_src,
		(hypptr->segs[hess+i].lm_s / ((int32) sc_ratio)),
		(((int32)(hypptr->segs[hess+i].acoust_s / num_frames)) /
		 ((int32) sc_ratio)));
      }
      else  {
	fprintf(tkfp,"tk_errwin_print_hyp_word %d %d %s %s %d %d\n",
		hypptr->segs[hess+i].start_f,
		hypptr->segs[hess+i].end_f,
		hypptr->segs[hess+i].word,
		hypptr->segs[hess+i].lm_s_src,
		(hypptr->segs[hess+i].lm_s / ((int32) sc_ratio)),
		0);
      }
      for (j=0;j<num_phones;j++)  {
	num_ph_frames = (int32) (hypptr->segs[hess+i].phone_seq.phones[j].ef -
				 hypptr->segs[hess+i].phone_seq.phones[j].sf);
	if (num_ph_frames > 0)  {
	  fprintf(tkfp,"tk_errwin_print_hyp_phone %d %d %s %d %d\n",
		  hypptr->segs[hess+i].phone_seq.phones[j].sf,
		  hypptr->segs[hess+i].phone_seq.phones[j].ef,
		  hypptr->segs[hess+i].phone_seq.phones[j].basename,
		  (((int32) (hypptr->segs[hess+i].phone_seq.phones[j].ac_score /
			     num_ph_frames)) /
		   ((int32) sc_ratio)),
		  hypptr->segs[hess+i].phone_seq.phones[j].correct);
	}
	else  {
	  fprintf(tkfp,"tk_errwin_print_hyp_phone %d %d %s %d %d\n",
		  hypptr->segs[hess+i].phone_seq.phones[j].sf,
		  hypptr->segs[hess+i].phone_seq.phones[j].ef,
		  hypptr->segs[hess+i].phone_seq.phones[j].basename,
		  0,
		  hypptr->segs[hess+i].phone_seq.phones[j].correct);
	}
      }
    }
    else  {
      fprintf(tkfp,"tk_errwin_print_hyp_word %d %d %s %s %d %d\n",
	      hypptr->segs[hess+i].start_f,
	      hypptr->segs[hess+i].end_f,
	      hypptr->segs[hess+i].word,
	      hypptr->segs[hess+i].lm_s_src,
	      (hypptr->segs[hess+i].lm_s / ((int32) sc_ratio)),
	      (hypptr->segs[hess+i].acoust_s / ((int32) sc_ratio)));
      for (j=0;j<num_phones;j++)  {
	fprintf(tkfp,"tk_errwin_print_hyp_phone %d %d %s %d %d\n",
		hypptr->segs[hess+i].phone_seq.phones[j].sf,
		hypptr->segs[hess+i].phone_seq.phones[j].ef,
		hypptr->segs[hess+i].phone_seq.phones[j].basename,
		(hypptr->segs[hess+i].phone_seq.phones[j].ac_score /
		 ((int32) sc_ratio)),
		hypptr->segs[hess+i].phone_seq.phones[j].correct);
      }
    }
  }

  fclose(tkfp);
}


/*  --------------  */
/*  Compares two objects of type force_out_seg.  Returns 1
    if word, start_f, and end_f are the same for the two segs.
    Returns 0 otherwise.
*/
int match_force_out_seg(ref,hyp)
force_out_seg_t *ref,*hyp;

{
#if 0
  printf("REF: %s (%d,%d)\n",
	 ref->word,ref->start_f,ref->end_f);

  printf("HYP: %s (%d,%d)\n",
	 hyp->word,hyp->start_f,hyp->end_f);
#endif
#if 0
  if ((!strcmp(ref->word,hyp->word)) &&
      (ref->start_f == hyp->start_f)  &&
      (ref->end_f == hyp->end_f))  {
    return(1);
  }
  else  {
    return(0);
  }
#endif
#if 0
  printf("ref start is %d\n",ref->start_f);
  printf("ref end   is %d\n",ref->end_f);
  printf("hyp start is %d\n",hyp->start_f);
  printf("hyp end   is %d\n",hyp->end_f);
  fflush(stdout);
#endif
#if 0
  printf("start diff is %d\n",
	 (abs((int)(ref->start_f - hyp->start_f))));
  printf("end diff is %d\n",
	 (abs((int)(ref->end_f - hyp->end_f))));
  fflush(stdout);
#endif
  if ((!strcmp(ref->word,hyp->word)) &&
      (abs((int)(ref->start_f - hyp->start_f)) <= frame_tol)  &&
      (abs((int)(ref->end_f - hyp->end_f)) <= frame_tol))  {
    return(1);
  }
  else  {
    return(0);
  }
}


/*********************************************/
/* Creates a new error region and returns a 
pointer to the newly created region. Also intializes
all fields. */

err_region_t *make_new_err_region(void)
{
  err_region_t *new_reg;

  new_reg = (err_region_t *) malloc(sizeof(err_region_t));
 
  if (new_reg == 0)  {
     printf("WARNING: Malloc failed when creating a new error region.\n");
     exit(-1);
  }

  new_reg->err_ID = 0;
  new_reg->ref = 0;
  new_reg->hyp = 0;
  new_reg->sf = 0;
  new_reg->ef = 0;
  new_reg->parent = 0;
  new_reg->num_children = 0;
  new_reg->child = 0;
  new_reg->ref_data.st_seg = 0;
  new_reg->ref_data.en_seg = 0;
  new_reg->hyp_data.st_seg = 0;
  new_reg->hyp_data.en_seg = 0;
  new_reg->ref_data.acoust_s = 0;
  new_reg->hyp_data.acoust_s = 0; 
  new_reg->ref_data.lm_s = 0;     
  new_reg->hyp_data.lm_s = 0;     
  new_reg->ref_data.tot_s = 0;    
  new_reg->hyp_data.tot_s = 0;    
  new_reg->ref_data.pruning = 0;
  new_reg->hyp_data.pruning = 0;
  new_reg->ref_data.align_type = 0;
  new_reg->hyp_data.align_type = 0;

  return (new_reg);
}      

