/*
 * lat.h - lattice dumping header file
 *
 * HISTORY
 *  created Feb. 28, 1995  Lin Chase (lindaq)
 *
 *  added LM and error analysis stuff April 5, 1995 Lin Chase
 *  moved error analysis stuff to error.h May 2, 1995 LLC
 */

#define MAX_UNDER_THRESH 20
#define ASTAR_MAXOUT 1000
#define BESTN_MAXSEG 256
#define MAX_OOV_WORDS 500
#define MAX_LAT_SEGS 1000
#define MAX_ALIGN_SEGS 200
#define MAX_ALIGN_PHONES 2000
#define MAX_WORD_PHONES 40
#define MAX_WORD_SEGS 1000
#define MAX_PREV_WDS 1000
#define MAX_LAT_PATHS 300
#define MAX_ALIGN_PATHS 30
#define LEFT 0
#define RIGHT 1
#define ALIGN_ENABLED 1
#define ALIGN_OFFSET 4

/* LM score types supported so far, actual values arbitrary*/
/* values of the global/shared LMSource variable  (see lat_read.c and lm_3g.c) */
#define T 11
#define BB_B 12
#define BB_BU 13
#define B 14
#define BU 15
#define U 16
/* for marking source of LM score of OOV words in gold standard alignments */
#define UNK 17
#define COMP 18
#define SILSrc 19
#define FillSrc 10

#define NO_SOURCE -1

/* 0 if <UNK> did not appear in the dictionary for the runs that
 *  created the lattice files being analyzed, 1 if it did.
 */
#define UNK_IN_DICT 1
#define COMP_IN_LM 0

#define CXT_DEP 1
#define CXT_INDEP 2
#define NO_TYPE -1

#define ALIGN_OFFSET 4
#define NO_IDX -1
/* NO_WORD = -1 is defined in dict.h */
/* #define SSIZE 20000 */  /* used in A* search */ /* defined in param.h */

typedef struct dump_segment_s {
  long int wid;		/* dict word-id,possibly alt pron */
  int sf;             /* start frame */
  int ef;             /* end frame */
  long int ac_score;     /* LOG acoustic score in context */
  long int lc_wid;      /* alt pron of left context word */
  long int rc_wid;      /* alt pron of right context word */
  int num_paths;     /* number of paths through this segment */
} dump_segment_t;

typedef struct phone_s {
  long int pid;        /* phone ID (which one, lin? -- ask RKM !!!) */
  int sf;         /* start frame of segmentation */
  int ef;         /* end frame of segmentation */
  long int ac_score; /* score of segmentation (presumably optimal state sequence) */
  char name[64];     /* string name of phone (really need to be this long?) */
  char basename[64]; /* string name of base phone */
  int err_reg;     /* 1 if in an error region, 0 if not */
  int correct;     /* 1 if HYP phone aligned properly with REF phone, 0 if not */
} phone_t;

typedef struct phone_seq_s {
  int num_phones; /* length of phone sequence */
  int phone_type; /* CXT_DEP or CXT_INDEP */
  phone_t phones[MAX_WORD_PHONES];  /* HACK!  Waste of space! */
  long int ac_score; /* sum of ac_score fields of component phones */
} phone_seq_t;

/* This has been heavily redefined in the port away from the CMU lattices.*/
typedef struct lat_segment_s {
  char word[128];	/* word string */
  int sf;               /* start frame */
  int ef;               /* end frame */
  long int ac_score;       /* LOG acoustic score in context */
  phone_seq_t phone_seq; /* phones in word */
} lat_segment_t;

/* ------- */
/* Used to capture structure of lattice while overlooking ac_score differences
    due to cross-word phone contexts */

typedef struct lat_wordseg_s {
  int sf;         /* start frame, same as in latsegs (below) */
  int ef;         /* end frame, same as in latsegs (below) */
  long int wid;      /* dict word-id, possibly an alt pron, same as in latsegs (below) */
  int numsegs;    /* number of lat_segs associated with this wordseg */
  lat_segment_t **latsegs;       /* ptr to ptr to first lat_segment */
  int lnum;      /* number of lat_wordsegs to left */
  struct lat_wordseg_s **lwords; /* ptr to first left word */
  int rnum;      /* number of lat_wordsegs to right */
  struct lat_wordseg_s **rwords; /* ptr to first right word */
} lat_wordseg_t;

/* ------- */
/* Used to make chains of lat_segs */

typedef struct lat_segs_ptr_s {
  int latseg_ID;  /* unique ID of 'from' node */
  int direction;  /* LEFT is 0, RIGHT is 1 */
  int num;        /* number of pointers */
  lat_segment_t **to_nodes; /* head of list of pointers of length 'num' */
} lat_segs_ptr_t;

/* This has been edited in the process of porting away from the CMU lattices. */
typedef struct lat_LMinfo_s {
  /*int latseg_ID;*/        /* what segment of parent_path does this correspond to? */
  lat_segment_t *latseg_ptr; /* same info, through pointer */
  char lm_src[128];           /* source of LM score */
  long int lm_score;       /* lm score, possibly simply SIL/fill insertion penalty */
  struct lat_path_s *parent_path; /* what path is this part of? */
} lat_LMinfo_t;
  
typedef struct lat_path_s {
  int path_ID;    /* ?? necessary ?? */
  int num_segs;   /* how many latsegs as components? */
  lat_segment_t **seg_ptrs;  /* component segments */
  lat_LMinfo_t **LM_ptrs;  /* LM scores corresponding to component segments */
  long int path_score; /* combined acoustic and LM */
} lat_path_t;


