/* MAIN.C - for ERA (Error Region Analysis)
 *
 * USAGE
 *	era <input_file> <output_file> <verbosity==1,2,or3>
 *
 * DESCRIPTION
 *	Read in a file containing segmentations of REF and HYP versions of utterances.
 *      Output error analysis based on error regions.
 *
 *-----------------------------------------------------------------------*
 * HISTORY
 *
 * 20-Jul-95  L L Chase (chase@cs.cmu.edu) at LM95/Johns Hopkins
 *      Extracted error analysis component from CMU lattice analysis toolkit.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <sys/param.h> 
#include <time.h>
#include <sys/resource.h> 
#include "lat.h"
#include "general.h"
#include "comp.h"
#include "error.h"

/* #include <c.h> */ /* ?? */


#define MAX_PATH_SEGS 128

/* from era.c */
extern int ref_tot_bigger;
extern int hyp_tot_bigger;
extern int num_error_regions;
extern int num_unk_error_regions;

extern int prune_regs_ac,no_prune_regs_ac;
extern int prune_regs_lm,no_prune_regs_lm;
extern int prune_regs_both,no_prune_regs_both;
extern float prune_both_ratio_lmtoa,no_prune_both_ratio_lmtoa;

void read_and_analyze_input(infp,window,tol,norm,tkdir,sc_ratio)
FILE *infp;
int window,tol,norm;
char *tkdir;
int sc_ratio;
{
  char Info[128],dummy[128];
  int sf=0,ef=0;
  int ph_sf=0,ph_ef=0;
  int i,j,k;
  char WordStr[128],utt_id_tag[128],PhoneStr[128];
  long int ac_score=0, lm_score=0;
  long int ph_ac_score=0, ph_lm_score=0;
  char lm_scoreSrc[128];
  int num_utts=0,num_segs=0;
  lat_path_t ref_latpath,hyp_latpath;
  lat_segment_t *ref_seg_ptrs[MAX_PATH_SEGS],*hyp_seg_ptrs[MAX_PATH_SEGS];
  lat_LMinfo_t *ref_LM_ptrs[MAX_PATH_SEGS],*hyp_LM_ptrs[MAX_PATH_SEGS];
  lat_segment_t *new_seg;
  lat_LMinfo_t *new_LMinfo;
  int num_phones = 0;

  Info[0] = '\0';
  dummy[0] = '\0';
  PhoneStr[0] = '\0';

  /*printf("Reading input file.\n");*/
  if (fgets(Info,120,infp) != NULL)  {
    sscanf(Info,"%d %s\n",&num_utts,dummy);
    if (!strcmp(dummy,"UTTERANCES"))  {
      if (num_utts <= 0)  {
	printf("WARNING: Input file contains zero (0) utterances.\n");
	exit(-1);
      }
      else  {
	/* printf("Reading utterance REF and HYP segmentations.\n");*/
      }
      for (i=0;i<num_utts;i++)  {
	/* Read Utterance tag */
	if (fgets(Info,120,infp) != NULL)  {
	  sscanf(Info,"%s %s\n",dummy,utt_id_tag);
	  if (!strcmp(dummy,"UTT"))  {
	    /*printf("Reading REF and HYP segmentations for utterance %s.\n",utt_id_tag);*/
	  }
	  else  {
	    printf("WARNING:  Input file format incorrect. (UTT #%d).\n",i);
	    exit(-1);
	  }
	}
	/* read REF segments */
	if (fgets(Info,120,infp) != NULL)  {
	  sscanf(Info,"%s\n",dummy);
	  if (!strcmp(dummy,"REF"))  {
	    /*printf("Reading REF segmentations.\n");*/
	    if (fgets(Info,120,infp) != NULL)  {
	      sscanf(Info,"%d %s\n",&num_segs,dummy);
	      if (!strcmp(dummy,"SEGMENTS"))  {
		/*printf("Reading %d REF segmentations.\n",num_segs);*/
		/* initialize ref_latpath data structure */
		ref_latpath.num_segs = 0;
		ref_latpath.seg_ptrs = NULL;
		ref_latpath.LM_ptrs = NULL;
		ref_latpath.path_score = 0;
		ref_latpath.path_ID = -1;
		for (j=0;j<MAX_PATH_SEGS;j++)  {
		  ref_seg_ptrs[j]=NULL;
		  ref_LM_ptrs[j]=NULL;
		}
		ref_latpath.seg_ptrs=ref_seg_ptrs;
		ref_latpath.LM_ptrs=ref_LM_ptrs;
		/* Read in num_segs REF segments. */
		for (j=0;j<num_segs;j++)  {
		  /* need to malloc ahead of time now that we're reading in 
		     phone segments */
		  new_seg = (lat_segment_t *) malloc(sizeof(lat_segment_t));
		  new_LMinfo = (lat_LMinfo_t *) malloc(sizeof(lat_LMinfo_t));
		  /* now read info from file */
		  if (fgets(Info,120,infp) != NULL)  {
		    sscanf(Info,"%s\n",WordStr);
		  }
		  else  {
		    printf("WARNING: Failure reading word, utt #%d at REF word #%d.\n",i,j);
		    exit(-1);
		  }
		  if (fgets(Info,120,infp) != NULL)  {
		    sscanf(Info,"%d\n",&sf);
		  }
		  else  {
		    printf("WARNING: Failure reading sf, utt #%d at REF word #%d.\n",i,j);
		    exit(-1);
		  }
		  if (fgets(Info,120,infp) != NULL)  {
		    sscanf(Info,"%d\n",&ef);
		  }
		  else  {
		    printf("WARNING: Failure reading ef, utt #%d at REF word #%d.\n",i,j);
		    exit(-1);
		  }
		  if (fgets(Info,120,infp) != NULL)  {
		    sscanf(Info,"%li\n",&ac_score);
		  }
		  else  {
		    printf("WARNING: Failure reading ac_score, utt #%d at REF word #%d.\n",i,j);
		    exit(-1);
		  }
		  if (fgets(Info,120,infp) != NULL)  {
		    sscanf(Info,"%li\n",&lm_score);
		  }
		  else  {
		    printf("WARNING: Failure reading lm_score, utt #%d at REF word #%d.\n",i,j);
		    exit(-1);
		  }
		  if (fgets(Info,120,infp) != NULL)  {
		    sscanf(Info,"%s\n",lm_scoreSrc);
		  }
		  else  {
		    printf("WARNING: Failure reading lm_score source, utt #%d at REF word #%d.\n",i,j);
		    exit(-1);
		  }
		  if (fgets(Info,120,infp) != NULL)  {
		    sscanf(Info,"%d\n",&num_phones);
		    new_seg->phone_seq.num_phones = num_phones;
		    for (k=0;k<num_phones;k++)  {
		      if (fgets(Info,120,infp) != NULL)  {
			sscanf(Info,"%s\n",PhoneStr);
			strcpy(new_seg->phone_seq.phones[k].name,PhoneStr);
		      }
		      else  {
			printf("WARNING: Failure reading phone string,utt #%d at REF word #%d.\n",i,j);
			exit(-1);
		      }
		      if (fgets(Info,120,infp) != NULL)  {
			sscanf(Info,"%d\n",&ph_sf);
			new_seg->phone_seq.phones[k].sf = ph_sf;
		      }
		      else  {
			printf("WARNING: Failure reading phone sf, utt #%d at REF word #%d.\n",i,j);
			exit(-1);
		      }
		      if (fgets(Info,120,infp) != NULL)  {
			sscanf(Info,"%d\n",&ph_ef);
			new_seg->phone_seq.phones[k].ef = ph_ef;
		      }
		      else  {
			printf("WARNING: Failure reading phone ef, utt #%d at REF word #%d.\n",i,j);
			exit(-1);
		      }
		      if (fgets(Info,120,infp) != NULL)  {
			sscanf(Info,"%li\n",&ph_ac_score);
			new_seg->phone_seq.phones[k].ac_score = ph_ac_score;
		      }
		      else  {
			printf("WARNING: Failure reading phone ac score, utt #%d at REF word #%d.\n",i,j);
			exit(-1);
		      }
		      /*printf("REF phone[%d] at segment[%d] is:\n",
			     k,j);
		      printf("%s (%d,%d) {ac:%li}\n",
			     new_seg->phone_seq.phones[k].name,
			     new_seg->phone_seq.phones[k].sf,
			     new_seg->phone_seq.phones[k].ef,
			     new_seg->phone_seq.phones[k].ac_score);
		      fflush(stdout);
		      */
		    }
		  }
		  else  {
		    printf("WARNING: Failure reading number of phones, utt #%d at REF word #%d.\n",i,j);
		    exit(-1);
		  }
		  /*printf("REF segment[%d] is:\n",j);
		  printf("%s (%d,%d) {ac:%li,lm:%li} <%s>\n",
			 WordStr,
			 sf,
			 ef,
			 ac_score,
			 lm_score,
			 lm_scoreSrc);
			 */
		  /* fill structure that will be passed to the error analysis routine */
		  ref_latpath.num_segs=num_segs;
/*		  new_seg = (lat_segment_t *) malloc(sizeof(lat_segment_t));*/
		  new_seg->word[0]='\0';
/*		  new_LMinfo = (lat_LMinfo_t *) malloc(sizeof(lat_LMinfo_t));*/
		  new_LMinfo->lm_src[0]='\0';
		  strcpy(new_seg->word,WordStr);
		  new_seg->sf = sf;
		  new_seg->ef = ef;
		  new_seg->ac_score = ac_score;
		  new_LMinfo->latseg_ptr = new_seg;
		  strcpy(new_LMinfo->lm_src,lm_scoreSrc);
		  new_LMinfo->lm_score = lm_score;
		  new_LMinfo->parent_path = &ref_latpath;
		  ref_latpath.seg_ptrs[j]=new_seg;
		  ref_latpath.LM_ptrs[j]=new_LMinfo;
		}
	      }
	      else  {
		printf("WARNING:  Input file format incorrect.  Cannot read number of REF segments at utt #%d.\n",i);
		exit(-1);
	      }
	    }
	  }
	  else  {
	    printf("WARNING:  Input file format incorrect. (REF at utt #%d.\n",i);
	    exit(-1);
	  }
	}
	/* end REF */
	/* read HYP segments */
	if (fgets(Info,120,infp) != NULL)  {
	  sscanf(Info,"%s\n",dummy);
	  if (!strcmp(dummy,"HYP"))  {
	    /*printf("Reading HYP segmentations.\n");*/
	    if (fgets(Info,120,infp) != NULL)  {
	      sscanf(Info,"%d %s\n",&num_segs,dummy);
	      if (!strcmp(dummy,"SEGMENTS"))  {
		/*printf("Reading %d HYP segmentations.\n",num_segs);*/
		/* initialize hyp_latpath data structure */
		hyp_latpath.num_segs = 0;
		hyp_latpath.seg_ptrs = NULL;
		hyp_latpath.LM_ptrs = NULL;
		hyp_latpath.path_score = 0;
		hyp_latpath.path_ID = -1;
		for (j=0;j<MAX_PATH_SEGS;j++)  {
		  hyp_seg_ptrs[j]=NULL;
		  hyp_LM_ptrs[j]=NULL;
		}
		hyp_latpath.seg_ptrs=hyp_seg_ptrs;
		hyp_latpath.LM_ptrs=hyp_LM_ptrs;
		/* Read in num_segs HYP segments. */
		for (j=0;j<num_segs;j++)  {
		  /* need to malloc ahead of time now that we're reading in 
		     phone segments */
		  new_seg = (lat_segment_t *) malloc(sizeof(lat_segment_t));
		  new_LMinfo = (lat_LMinfo_t *) malloc(sizeof(lat_LMinfo_t));
		  /* now read info from file */
		  if (fgets(Info,120,infp) != NULL)  {
		    sscanf(Info,"%s\n",WordStr);
		  }
		  else  {
		    printf("WARNING: Failure reading word, utt #%d at HYP word #%d.\n",i,j);
		    exit(-1);
		  }
		  if (fgets(Info,120,infp) != NULL)  {
		    sscanf(Info,"%d\n",&sf);
		  }
		  else  {
		    printf("WARNING: Failure reading sf, utt #%d at HYP word #%d.\n",i,j);
		    exit(-1);
		  }
		  if (fgets(Info,120,infp) != NULL)  {
		    sscanf(Info,"%d\n",&ef);
		  }
		  else  {
		    printf("WARNING: Failure reading ef, utt #%d at HYP word #%d.\n",i,j);
		    exit(-1);
		  }
		  if (fgets(Info,120,infp) != NULL)  {
		    sscanf(Info,"%li\n",&ac_score);
		  }
		  else  {
		    printf("WARNING: Failure reading ac_score, utt #%d at HYP word #%d.\n",i,j);
		    exit(-1);
		  }
		  if (fgets(Info,120,infp) != NULL)  {
		    sscanf(Info,"%li\n",&lm_score);
		  }
		  else  {
		    printf("WARNING: Failure reading lm_score, utt #%d at HYP word #%d.\n",i,j);
		    exit(-1);
		  }
		  if (fgets(Info,120,infp) != NULL)  {
		    sscanf(Info,"%s\n",lm_scoreSrc);
		  }
		  else  {
		    printf("WARNING: Failure reading lm_score source, utt #%d at HYP word #%d.\n",i,j);
		    exit(-1);
		  }
		  if (fgets(Info,120,infp) != NULL)  {
		    sscanf(Info,"%d\n",&num_phones);
		    new_seg->phone_seq.num_phones = num_phones;
		    for (k=0;k<num_phones;k++)  {
		      if (fgets(Info,120,infp) != NULL)  {
			sscanf(Info,"%s\n",PhoneStr);
			strcpy(new_seg->phone_seq.phones[k].name,PhoneStr);
		      }
		      else  {
			printf("WARNING: Failure reading phone string,utt #%d at REF word #%d.\n",i,j);
			exit(-1);
		      }
		      if (fgets(Info,120,infp) != NULL)  {
			sscanf(Info,"%d\n",&ph_sf);
			new_seg->phone_seq.phones[k].sf = ph_sf;
		      }
		      else  {
			printf("WARNING: Failure reading phone sf, utt #%d at REF word #%d.\n",i,j);
			exit(-1);
		      }
		      if (fgets(Info,120,infp) != NULL)  {
			sscanf(Info,"%d\n",&ph_ef);
			new_seg->phone_seq.phones[k].ef = ph_ef;
		      }
		      else  {
			printf("WARNING: Failure reading phone ef, utt #%d at REF word #%d.\n",i,j);
			exit(-1);
		      }
		      if (fgets(Info,120,infp) != NULL)  {
			sscanf(Info,"%li\n",&ph_ac_score);
			new_seg->phone_seq.phones[k].ac_score = ph_ac_score;
		      }
		      else  {
			printf("WARNING: Failure reading phone ac score, utt #%d at REF word #%d.\n",i,j);
			exit(-1);
		      }
		      /*printf("HYP phone[%d] at segment[%d] is:\n",
			     k,j);
		      printf("%s (%d,%d) {ac:%li}\n",
			     new_seg->phone_seq.phones[k].name,
			     new_seg->phone_seq.phones[k].sf,
			     new_seg->phone_seq.phones[k].ef,
			     new_seg->phone_seq.phones[k].ac_score);
		      fflush(stdout);
		      */
		    }
		  }
		  else  {
		    printf("WARNING: Failure reading number of phones, utt #%d at REF word #%d.\n",i,j);
		    exit(-1);
		  }
		  /*printf("HYP segment[%d] is:\n",j);
		  printf("%s (%d,%d) {ac:%li,lm:%li} <%s>\n",
			 WordStr,
			 sf,
			 ef,
			 ac_score,
			 lm_score,
			 lm_scoreSrc);
			 */
		  /* fill structure that will be passed to the error analysis routine */
		  hyp_latpath.num_segs=num_segs;
/*		  new_seg = (lat_segment_t *) malloc(sizeof(lat_segment_t));*/
		  new_seg->word[0]='\0';
/*		  new_LMinfo = (lat_LMinfo_t *) malloc(sizeof(lat_LMinfo_t));*/
		  new_LMinfo->lm_src[0]='\0';
		  strcpy(new_seg->word,WordStr);
		  new_seg->sf = sf;
		  new_seg->ef = ef;
		  new_seg->ac_score = ac_score;
		  new_LMinfo->latseg_ptr = new_seg;
		  strcpy(new_LMinfo->lm_src,lm_scoreSrc);
		  new_LMinfo->lm_score = lm_score;
		  new_LMinfo->parent_path = &hyp_latpath;
		  hyp_latpath.seg_ptrs[j]=new_seg;
		  hyp_latpath.LM_ptrs[j]=new_LMinfo;
		}
	      }
	      else  {
		printf("WARNING:  Input file format incorrect.  Cannot read number of HYP segments at utt #%d.\n",i);
		exit(-1);
	      }
	    }
	  }
	  else  {
	    printf("WARNING:  Input file format incorrect. (HYP at utt #%d.)\n",i);
	    exit(-1);
	  }
	}
	/* end HYP */
	/*printf("Read utterance [%d] %s successfully.\n",i,utt_id_tag);*/
	printf("Utterance %s\n",utt_id_tag);
	do_error_reg_analysis(&ref_latpath,
			      &hyp_latpath,
			      utt_id_tag,
			      tkdir,
			      window,
			      tol,
			      norm,
			      sc_ratio);
      }
    }
    else  {
      printf("WARNING:  Input file format incorrect. (UTTERANCES).\n");
      exit(-1);
    }
  }
}

void main (argc, argv)
/*-------------------------------------------------------*/
int argc;
char **argv;
{
  int window = 0;
  FILE *infp;
  int frame_tol = 0;
  int normalize = 0;
  int sc_ratio = 1000;

  if (argc != 7)  {
    printf("USAGE: visera <input_file> <window==0,1,or2> <frame tolerance==n> <normalize AC scores==0,1 <output directory> <score ratio>\n");
    exit(-1);
  }
  else  {
    ref_tot_bigger = 0;
    hyp_tot_bigger = 0;

    prune_regs_ac = 0;
    no_prune_regs_ac = 0;

    prune_regs_lm = 0;
    no_prune_regs_lm = 0;

    prune_regs_both = 0;
    no_prune_regs_both = 0;

    prune_both_ratio_lmtoa = 1.0;
    no_prune_both_ratio_lmtoa = 1.0;

    num_error_regions = 0;
    num_unk_error_regions = 0;

    window = atoi(argv[2]);
    frame_tol = atoi(argv[3]);
    normalize = atoi(argv[4]);
    sc_ratio = atoi(argv[6]);
    if ((frame_tol > 0) && (normalize != 1))  {
      printf("If you use a frame tolerance greater than zero you must report normalized acoustic scores.  Use the value normalize == 1 as input.\n");
      exit(-2);
    }
    /*printf ("\n%s:\n input file %s\n window ==%d\n tolerance ==%d\n normalizing AC scores ==%d score ratio ==%d\n",
	    argv[0], argv[1], window, frame_tol, normalize, sc_ratio);
	    */
    infp = fopen (argv[1],"r");
    if (NULL == infp)  {
      printf("Problem opening infile.  Exiting.\n");
      exit(-1);
    }
    if (sc_ratio <= 0)  {
      printf("ERROR.  sc_ratio must be an integer greater than 0, preferably a multiple of ten such as 100 or 1000.\n");
      exit(-1);
    }
    read_and_analyze_input(infp,window,frame_tol,normalize,argv[5],sc_ratio);
    if ( 0 != fclose(infp))  {
      printf("Problem closing infile.  Exiting.\n");
      exit(-1);
    }
    /*printf("ErrorRegions %d\n",num_error_regions);*/
    printf("UNKErrRegs %d\n",num_unk_error_regions);
    printf("PruningErrRegs %d \n",ref_tot_bigger);
    printf("PruneAC %d\n",prune_regs_ac);
    printf("PruneLM %d\n",prune_regs_lm);
    printf("PruneBo %d\n",prune_regs_both);
#if 0
    if (prune_regs_both>0)  {
      printf("PruneBoRatio %3.0f\n",(float)(prune_both_ratio_lmtoa/prune_regs_both));
    }
    else  {
      printf("PruneNoRatio\n");
    }
#endif
    printf("ModelingErrRegs %d \n",hyp_tot_bigger);
    printf("ModelAC %d\n",no_prune_regs_ac);
    printf("ModelLM %d\n",no_prune_regs_lm);
    printf("ModelBo %d\n",no_prune_regs_both);
#if 0
    if (no_prune_regs_both>0)  {
      printf("ModelBoRatio  %3.0f\n",(float)(no_prune_both_ratio_lmtoa/no_prune_regs_both));
    }
    else  {
      printf("ModelNoRatio\n");
    }
#endif
  }
}


